# -*- coding: utf-8 -*-

from collections import OrderedDict, defaultdict
from functools import reduce
from itertools import chain, product
from operator import mul, add
import copy
import math
import warnings

from .util.arithmeticdict import ArithmeticDict
from .util._expr import Expr
from .util.periodic import mass_from_composition
from .util.parsing import (
    formula_to_composition,
    to_reaction,
    formula_to_latex,
    formula_to_unicode,
    formula_to_html,
)

from .units import default_units, is_quantity, unit_of, to_unitless
from ._util import intdiv
from .util.pyutil import deprecated, DeferredImport, ChemPyDeprecationWarning


ReactionSystem = DeferredImport(
    "chempy.reactionsystem",
    "ReactionSystem",
    [deprecated(use_instead="chempy.ReactionSystem")],
)


class Substance(object):
    """ Class representing a chemical substance

    Parameters
    ----------
    name : str
    charge : int (optional, default: None)
        Will be stored in composition[0], prefer composition when possible.
    latex_name : str
    unicode_name : str
    html_name : str
    composition : dict or None (default)
        Dictionary (int -> number) e.g. {atomic number: count}, zero has special
        meaning (net charge). Avoid using the key 0 unless you specifically mean
        net charge. The motivation behind this is that it is easier to track a
        net-charge of e.g. 6 for U(VI) than it is to remember that uranium has 92
        electrons and use 86 as the value).
    data : dict
        Free form dictionary. Could be simple such as ``{'mp': 0, 'bp': 100}``
        or considerably more involved, e.g.: ``{'diffusion_coefficient': {\
 'water': lambda T: 2.1*m**2/s/K*(T - 273.15*K)}}``.

    Attributes
    ----------
    mass
        Maps to data['mass'], and when unavailable looks for ``formula.mass``.
    attrs
        A tuple of attribute names for serialization.
    composition : dict or None
        Dictionary mapping fragment key (str) to amount (int).
    data
        Free form dictionary.

    Examples
    --------
    >>> ammonium = Substance('NH4+', 1, 'NH_4^+', composition={7: 1, 1: 4},
    ...     data={'mass': 18.0385, 'pKa': 9.24})
    >>> ammonium.name
    'NH4+'
    >>> ammonium.composition == {0: 1, 1: 4, 7: 1}  # charge represented by key '0'
    True
    >>> ammonium.data['mass']
    18.0385
    >>> ammonium.data['pKa']
    9.24
    >>> ammonium.mass  # mass is a special case (also attribute)
    18.0385
    >>> ammonium.pKa
    Traceback (most recent call last):
        ...
    AttributeError: 'Substance' object has no attribute 'pKa'
    >>> nh4p = Substance.from_formula('NH4+')  # simpler
    >>> nh4p.composition == {7: 1, 1: 4, 0: 1}
    True
    >>> nh4p.latex_name
    'NH_{4}^{+}'

    """

    attrs = ("name", "latex_name", "unicode_name", "html_name", "composition", "data")

    def __eq__(self, other):
        for attr in self.attrs:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __hash__(self) -> int:
        hashed_values = []
        for key in self.attrs:
            value = getattr(self, key)
            if isinstance(value, dict):
                hashed_values.append(hash(tuple(sorted(value.items()))))
            else:
                hashed_values.append(hash(value))
        return sum(hashed_values)

    @property
    def charge(self):
        """Convenience property for accessing ``composition[0]``"""
        return self.composition.get(0, 0)  # electron (net) deficiency

    @property
    def mass(self):
        """Convenience property for accessing ``data['mass']``

        when ``data['mass']`` is missing the mass is calculated
        from the :attr:`composition` using
        :func:`chempy.util.parsing.mass_from_composition`.
        """
        try:
            return self.data["mass"]
        except KeyError:
            if self.composition is not None:
                return mass_from_composition(self.composition)

    @mass.setter
    def mass(self, value):
        self.data["mass"] = value

    def molar_mass(self, units=None):
        """Returns the molar mass (with units) of the substance

        Examples
        --------
        >>> nh4p = Substance.from_formula('NH4+')  # simpler
        >>> from chempy.units import default_units as u
        >>> nh4p.molar_mass(u)
        array(18.0384511...) * g/mol

        """
        if units is None:
            units = default_units
        return self.mass * units.g / units.mol

    def __init__(
        self,
        name=None,
        charge=None,
        latex_name=None,
        unicode_name=None,
        html_name=None,
        composition=None,
        data=None,
    ):
        self.name = name
        self.latex_name = latex_name
        self.unicode_name = unicode_name
        self.html_name = html_name
        self.composition = composition

        if self.composition is not None and 0 in self.composition:
            if charge is not None:
                raise KeyError("Cannot give both charge and composition[0]")
        else:
            if charge is not None and composition is not None:
                self.composition[0] = charge
        self.data = data or {}

    @classmethod
    def from_formula(cls, formula, **kwargs):
        """Creates a :class:`Substance` instance from its formula

        Parameters
        ----------
        formula: str
            e.g. 'Na+', 'H2O', 'Fe(CN)6-4'
        \\*\\*kwargs:
            keyword arguments passed on to `.Substance`

        Examples
        --------
        >>> NH3 = Substance.from_formula('NH3')
        >>> NH3.composition == {1: 3, 7: 1}
        True
        >>> '%.2f' % NH3.mass
        '17.03'
        >>> NH3.charge
        0
        >>> NH3.latex_name
        'NH_{3}'

        """
        return cls(
            formula,
            latex_name=formula_to_latex(formula),
            unicode_name=formula_to_unicode(formula),
            html_name=formula_to_html(formula),
            composition=formula_to_composition(formula),
            **kwargs
        )

    def __repr__(self):
        kw = ["name=" + self.name + ", ..."]  # Too verbose to print all
        return "<{}({})>".format(self.__class__.__name__, ",".join(kw))

    def __str__(self):
        return str(self.name)

    def _repr_html_(self):
        return self.html_name

    @staticmethod
    def composition_keys(substance_iter, skip_keys=()):
        """Occurring :attr:`composition` keys among a series of substances"""
        keys = set()
        for s in substance_iter:
            if s.composition is None:
                continue
            for k in s.composition.keys():
                if k in skip_keys:
                    continue
                keys.add(k)
        return sorted(keys)


class Species(Substance):
    """Substance belonging to a phase

    Species extends :class:`Substance` with the new attribute :attr:`phase_idx`

    Attributes
    ----------
    phase_idx: int
        Index of the phase (default is 0)
    """

    def __init__(self, *args, **kwargs):
        phase_idx = kwargs.pop("phase_idx", 0)
        super(Species, self).__init__(*args, **kwargs)
        self.phase_idx = phase_idx

    @property
    @deprecated(last_supported_version="0.3.0", will_be_missing_in="0.8.0")
    def precipitate(self):
        """deprecated attribute, provided for compatibility for now"""
        return self.phase_idx > 0

    @classmethod
    def from_formula(
        cls, formula, phases=("(s)", "(l)", "(g)"), default_phase_idx=0, **kwargs
    ):
        """Create a :class:`Species` instance from its formula

        Analogous to :meth:`Substance.from_formula` but with the addition that
        phase_idx is determined from the formula (and a mapping provided by
        ``phases``)

        Parameters
        ----------
        formula: str
            e.g. 'H2O', 'NaCl(s)', 'CO2(aq)', 'CO2(g)'
        phases: iterable of str or dict mapping str -> int
            if not in \\*\\*kwargs, ``phase_idx`` is determined from the suffix
            of ``formula`` where the suffixes is mapped from phases:
                if ``phases`` is a dictionary:
                    ``phase_idx = phases[suffix]``
                else:
                    ``phase_idx = phases.index(suffix) + 1``
            and if suffixes is missing in phases phase_idx is taken to be 0
        default_phase_idx: int or None (default: 0)
            If ``default_phase_idx`` is ``None``, ``ValueError`` is raised for
                unknown suffixes.
            Else ``default_phase_idx`` is used as ``phase_idx`` in those cases.
        \\*\\*kwargs:
            Keyword arguments passed on.

        Examples
        --------
        >>> water = Species.from_formula('H2O')
        >>> water.phase_idx
        0
        >>> NaCl = Species.from_formula('NaCl(s)')
        >>> NaCl.phase_idx
        1
        >>> Hg_l = Species.from_formula('Hg(l)')
        >>> Hg_l.phase_idx
        2
        >>> CO2g = Species.from_formula('CO2(g)')
        >>> CO2g.phase_idx
        3
        >>> CO2aq = Species.from_formula('CO2(aq)', default_phase_idx=None)
        Traceback (most recent call last):
            ...
        ValueError: Could not determine phase_idx
        >>> CO2aq = Species.from_formula('CO2(aq)')
        >>> CO2aq.phase_idx
        0
        >>> CO2aq = Species.from_formula('CO2(aq)', ['(aq)'],
        ...     default_phase_idx=None)
        >>> CO2aq.phase_idx
        1
        >>> Species.from_formula('CO2(aq)', {'(aq)': 0}, None).phase_idx
        0


        Raises
        ------
        ValueError:
            if ``default_phase_idx`` is ``None`` and no suffix found in phases

        """
        if "phase_idx" in kwargs:
            p_i = kwargs.pop("phase_idx")
        else:
            p_i = None
            if isinstance(phases, dict):
                for k, v in phases.items():
                    if formula.endswith(k):
                        p_i = v
                        break
            else:
                for idx, phase in enumerate(phases):
                    if formula.endswith(phase):
                        p_i = idx + 1
                        break
            if p_i is None:
                if default_phase_idx is None:
                    raise ValueError("Could not determine phase_idx")
                else:
                    p_i = default_phase_idx
        return cls(
            formula,
            latex_name=formula_to_latex(formula, suffixes=phases),
            unicode_name=formula_to_unicode(formula, suffixes=phases),
            html_name=formula_to_html(formula, suffixes=phases),
            composition=formula_to_composition(formula, suffixes=phases),
            phase_idx=p_i,
            **kwargs
        )


@deprecated(
    last_supported_version="0.3.0", will_be_missing_in="0.8.0", use_instead=Species
)
class Solute(Substance):
    """[DEPRECATED] Use `.Species` instead

    Counter-intuitive to its name Solute has an additional
    property 'precipitate'

    """

    def __init__(self, *args, **kwargs):
        precipitate = kwargs.pop("precipitate", False)
        Substance.__init__(self, *args, **kwargs)
        self.precipitate = precipitate

    @classmethod
    def from_formula(cls, formula, **kwargs):
        if formula.endswith("(s)"):
            kwargs["precipitate"] = True
        return cls(
            formula,
            latex_name=formula_to_latex(formula),
            unicode_name=formula_to_unicode(formula),
            html_name=formula_to_html(formula),
            composition=formula_to_composition(formula),
            **kwargs
        )


class Reaction(object):
    """Class representing a chemical reaction

    Consider for example:

        2 R --> A + P; r = k*A*R*R

    this would be represented as ``Reaction({'A': 1, 'R': 2},
    {'A': 2, 'P': 1}, param=k)``. Some reactions have a larger
    stoichiometric coefficient than what appears in the rate
    expression, e.g.:

        5 A + B --> C; r = k*A*B

    this can be represented as ``Reaction({'C1': 1, 'C2': 1},
    {'B': 1}, inact_reac={'C1': 4}, param=k)``.

    The rate constant information in ``param`` may be a subclass of
    :class:`chempy.kinetics.rates.RateExpr` or carry a :meth:`as_RateExpr`,
    if neither: `param` will be assumed to be a rate constant for a mass-action
    type of kinetic expression.

    Additional data may be stored in the ``data`` dict.


    Parameters
    ----------
    reac : dict (str -> int)
        If ``reac`` is a ``set``, then multiplicities are assumed to be 1.
    prod : dict (str -> int)
        If ``prod`` is a ``set``, then multiplicities are assumed to be 1.
    param : float or callable
        Special case (side-effect): if param is a subclass of
        :class:`.kinetics.rates.RateExpr` and its :attr:`rxn`
        is `None` it will be set to `self`.
    inact_reac : dict (optional)
    inact_prod : dict (optional)
    name : str (optional)
    k : deprecated (alias for param)
    ref : object
        Reference (e.g. a string containing doi number).
    data : dict (optional)
    checks : iterable of str
        Raises ``ValueError`` if any method ``check_%s`` returns False
        for all ``%s`` in ``checks``. Default: ``Reaction.default_checks``.

    Attributes
    ----------
    reac : OrderedDict
    prod : OrderedDict
    param : object
    inact_reac : OrderedDict
    inact_prod : OrderedDict
    name : str
    ref : str
    data : dict

    Examples
    --------
    >>> r = Reaction({'H2': 2, 'O2': 1}, {'H2O': 2})
    >>> r.keys() == {'H2', 'O2', 'H2O'}
    True
    >>> r.order()
    3
    >>> r.net_stoich(['H2', 'H2O', 'O2'])
    (-2, 2, -1)
    >>> print(r)
    2 H2 + O2 -> 2 H2O

    """

    _cmp_attr = ("reac", "prod", "param", "inact_reac", "inact_prod")
    _all_attr = _cmp_attr + ("name", "ref", "data")
    _str_arrow = "->"

    param_char = "k"  # convention
    default_checks = {"any_effect", "all_positive", "all_integral", "consistent_units"}

    @staticmethod
    def _init_stoich(container):
        if isinstance(container, set):
            container = {k: 1 for k in container}
        container = container or {}
        if type(container) == dict:  # noqa
            # we don't want isinstance here in case of OrderedDict
            container = OrderedDict(sorted(container.items(), key=lambda kv: kv[0]))
        return container

    def __init__(
        self,
        reac,
        prod,
        param=None,
        inact_reac=None,
        inact_prod=None,
        name=None,
        ref=None,
        data=None,
        checks=None,
        dont_check=None,
    ):
        self.reac = self._init_stoich(reac)
        self.inact_reac = self._init_stoich(inact_reac)
        self.prod = self._init_stoich(prod)
        self.inact_prod = self._init_stoich(inact_prod)
        self.param = param
        self.name = name
        self.ref = ref
        self.data = data or {}
        if checks is not None and dont_check is not None:
            raise ValueError("Cannot specify both checks and dont_check")
        if checks is None:
            checks = self.default_checks ^ (dont_check or set())

        for check in checks:
            getattr(self, "check_" + check)(throw=True)

    @classmethod
    def from_string(cls, string, substance_keys=None, globals_=None, **kwargs):
        """Parses a string into a Reaction instance

        Parameters
        ----------
        string : str
            String representation of the reaction.
        substance_keys : convertible to iterable of strings or string or None
            Used prevent e.g. misspelling.
            if str: split is invoked, if None: no checking done.
        globals_ : dict (optional)
            Dictionary for eval for (default: None -> {'chempy': chempy})
            If ``False``: no eval will be called (useful for web-apps).
        \\*\\*kwargs :
            Passed on to constructor.

        Examples
        --------
        >>> r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", 'H2O H+ OH-')
        >>> r.reac == {'H2O': 1} and r.prod == {'H+': 1, 'OH-': 1}
        True
        >>> r2 = Reaction.from_string("2 H2O -> 2 H2 + O2", 'H2O H2 O2')
        >>> r2.reac == {'H2O': 2} and r2.prod == {'H2': 2, 'O2': 1}
        True
        >>> r3 = Reaction.from_string("A -> B; 1/second", 'A B')
        >>> from chempy.units import to_unitless, default_units as u
        >>> to_unitless(r3.param, u.hour**-1).item()
        3600.0
        >>> r4 = Reaction.from_string("A -> 2 B; 'k'", 'A B')
        >>> r4.rate(dict(A=3, B=5, k=7)) == {'A': -3*7, 'B': 2*3*7}
        True
        >>> r5 = Reaction.from_string("A -> B; 1/molar/second", 'A B')  # DOCTEST: +ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: Incompatible units


        Notes
        -----
        :func:`chempy.util.parsing.to_reaction` is used which in turn calls
        :func:`eval` which is a severe security concern for untrusted input.

        """
        if isinstance(substance_keys, str):
            if " " in substance_keys:
                substance_keys = substance_keys.split()
        return to_reaction(
            string, substance_keys, cls._str_arrow, cls, globals_, **kwargs
        )

    def copy(self, **kwargs):
        if "checks" not in kwargs:
            kwargs["checks"] = ()
        for k in self._all_attr:
            if k not in kwargs:
                kwargs[k] = copy.copy(getattr(self, k))
        return self.__class__(**kwargs)

    def check_any_effect(self, throw=False):
        """Checks if the reaction has any effect"""
        if not any(self.net_stoich(self.keys())):
            if throw:
                raise ValueError(
                    "The net stoichiometry change of all species are zero."
                )
            else:
                return False
        return True

    def check_all_positive(self, throw=False):
        """Checks if all stoichiometric coefficients are positive"""
        for nam, cont in [
            (nam, getattr(self, nam))
            for nam in "reac prod inact_reac inact_prod".split()
        ]:
            for k, v in cont.items():
                if v < 0:
                    if throw:
                        raise ValueError(
                            "Found a negative stoichiometry for %s in %s." % (k, nam)
                        )
                    else:
                        return False
        return True

    def check_all_integral(self, throw=False):
        """Checks if all stoichiometric coefficients are integers"""
        for nam, cont in [
            (nam, getattr(self, nam))
            for nam in "reac prod inact_reac inact_prod".split()
        ]:
            for k, v in cont.items():
                if v != int(v) and v != type(v)(int(v)):
                    if throw:
                        raise ValueError(
                            "Found a non-integer stoichiometric coefficient for %s in %s."
                            % (k, nam)
                        )
                    else:
                        return False
        return True

    def check_consistent_units(self, throw=False):
        if is_quantity(self.param):  # This will assume mass action
            if isinstance(self.param.item(), Expr):
                param = (
                    self.param.item()({"temperature": 1 * default_units.K})
                    * self.param.units
                )
            else:
                param = self.param
            try:
                to_unitless(
                    param
                    / (default_units.molar ** (1 - self.order()) / default_units.s)
                )
            except Exception:
                if throw:
                    raise
                else:
                    return False
            else:
                return True
        else:
            return True  # the user might not be using ``chempy.units``

    def __eq__(lhs, rhs):
        if lhs is rhs:
            return True
        if not isinstance(lhs, Reaction) or not isinstance(rhs, Reaction):
            return NotImplemented
        for attr in lhs._cmp_attr:
            if getattr(lhs, attr) != getattr(rhs, attr):
                return False
        return True

    def __hash__(self):
        return sum(
            map(
                hash,
                (
                    getattr(self, k)
                    for k in ["reac", "prod", "param", "inact_reac", "inact_prod"]
                ),
            )
        )

    def order(self):
        """Sum of (active) reactant stoichiometries"""
        return sum(self.reac.values())

    def keys(self):
        return set(
            chain(
                self.reac.keys(),
                self.prod.keys(),
                self.inact_reac.keys(),
                self.inact_prod.keys(),
            )
        )

    def net_stoich(self, substance_keys):
        """Per substance net stoichiometry tuple (active & inactive)"""
        return tuple(
            self.prod.get(k, 0)
            - self.reac.get(k, 0)
            + self.inact_prod.get(k, 0)
            - self.inact_reac.get(k, 0)
            for k in substance_keys
        )

    def all_reac_stoich(self, substances):
        """Per substance reactant stoichiometry tuple (active & inactive)"""
        return tuple(
            self.reac.get(k, 0) + self.inact_reac.get(k, 0) for k in substances
        )

    def active_reac_stoich(self, substances):
        """Per substance reactant stoichiometry tuple (active)"""
        return tuple(self.reac.get(k, 0) for k in substances)

    def all_prod_stoich(self, substances):
        """Per substance product stoichiometry tuple (active & inactive)"""
        return tuple(
            self.prod.get(k, 0) + self.inact_prod.get(k, 0) for k in substances
        )

    def active_prod_stoich(self, substances):
        """Per substance product stoichiometry tuple (active)"""
        return tuple(self.prod.get(k, 0) for k in substances)

    def _xprecipitate_stoich(self, substances, xor):
        return tuple(
            (
                0
                if xor ^ (getattr(v, "phase_idx", 0) > 0)
                else self.prod.get(k, 0)
                + self.inact_prod.get(k, 0)
                - self.reac.get(k, 0)
                - self.inact_reac.get(k, 0)
            )
            for k, v in substances.items()
        )

    def precipitate_stoich(self, substances):
        """Only stoichiometry of precipitates"""
        net = self._xprecipitate_stoich(substances, True)
        found1 = -1
        for idx in range(len(net)):
            if net[idx] != 0:
                if found1 == -1:
                    found1 = idx
                else:
                    raise NotImplementedError("Only one precipitate assumed.")
        return net, net[found1], found1

    def non_precipitate_stoich(self, substances):
        """Only stoichiometry of non-precipitates"""
        return self._xprecipitate_stoich(substances, False)

    def has_precipitates(self, substances):
        for s_name in chain(
            self.reac.keys(),
            self.prod.keys(),
            self.inact_reac.keys(),
            self.inact_prod.keys(),
        ):
            if getattr(substances[s_name], "phase_idx", 0) > 0:
                return True
        return False

    def string(self, substances=None, with_param=False, with_name=False, **kwargs):
        """Returns a string representation of the reaction

        Parameters
        ----------
        substances: dict
            mapping substance keys to Substance instances
        with_param: bool
            whether to print the parameter (default: False)
        with_name: bool
            whether to print the name (default: False)

        Examples
        --------
        >>> r = Reaction({'H+': 1, 'Cl-': 1}, {'HCl': 1}, 1e10)
        >>> r.string(with_param=False)
        'Cl- + H+ -> HCl'

        """
        from .printing import str_

        return str_(
            self,
            substances=substances,
            with_param=with_param,
            with_name=with_name,
            **kwargs
        )

    def __str__(self):
        return self.string(with_param=True, with_name=True)

    def latex(self, substances, with_param=False, with_name=False, **kwargs):
        r"""Returns a LaTeX representation of the reaction

        Parameters
        ----------
        substances: dict
            mapping substance keys to Substance instances
        with_param: bool
            whether to print the parameter (default: False)
        with_name: bool
            whether to print the name (default: False)

        Examples
        --------
        >>> keys = 'H2O H+ OH-'.split()
        >>> subst = {k: Substance.from_formula(k) for k in keys}
        >>> r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", subst)
        >>> r.latex(subst) == r'H_{2}O \rightarrow H^{+} + OH^{-}'
        True
        >>> r2 = Reaction.from_string("H+ + OH- -> H2O; 1e8/molar/second", subst)
        >>> ref = r'H^{+} + OH^{-} \rightarrow H_{2}O; 10^{8} $\mathrm{\frac{1}{(s{\cdot}M)}}$'
        >>> r2.latex(subst, with_param=True) == ref
        True

        """
        from .printing import latex

        return latex(
            self,
            substances=substances,
            with_param=with_param,
            with_name=with_name,
            **kwargs
        )

    def unicode(self, substances, with_param=False, with_name=False, **kwargs):
        u"""Returns a unicode string representation of the reaction

        Examples
        --------
        >>> keys = 'H2O H+ OH-'.split()
        >>> subst = {k: Substance.from_formula(k) for k in keys}
        >>> r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", subst)
        >>> r.unicode(subst) == u'H₂O → H⁺ + OH⁻'
        True
        >>> r2 = Reaction.from_string("H+ + OH- -> H2O; 1e8/molar/second", subst)
        >>> r2.unicode(subst, with_param=True) == u'H⁺ + OH⁻ → H₂O; 10⁸ 1/(s·M)'
        True

        """
        from .printing import unicode_

        return unicode_(
            self,
            substances=substances,
            with_param=with_param,
            with_name=with_name,
            **kwargs
        )

    def html(self, substances, with_param=False, with_name=False, **kwargs):
        """Returns a HTML representation of the reaction

        Examples
        --------
        >>> keys = 'H2O H+ OH-'.split()
        >>> subst = {k: Substance.from_formula(k) for k in keys}
        >>> r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", subst)
        >>> r.html(subst)
        'H<sub>2</sub>O &rarr; H<sup>+</sup> + OH<sup>-</sup>'
        >>> r2 = Reaction.from_string("H+ + OH- -> H2O; 1e8/molar/second", subst)
        >>> r2.html(subst, with_param=True)
        'H<sup>+</sup> + OH<sup>-</sup> &rarr; H<sub>2</sub>O&#59; 10<sup>8</sup> 1/(s*M)'

        """
        from .printing import html

        return html(
            self,
            with_param=with_param,
            with_name=with_name,
            substances=substances,
            **kwargs
        )

    def _repr_html_(self):
        return self.html({k: k for k in self.keys()})

    def _violation(self, substances, attr):
        net = 0.0
        for substance, coeff in zip(
            substances.values(), self.net_stoich(substances.keys())
        ):
            net += getattr(substance, attr) * coeff
        return net

    def mass_balance_violation(self, substances):
        """Net amount of mass produced

        Parameters
        ----------
        substances: dict

        Returns
        -------
        float: amount of net mass produced/consumed

        """
        return self._violation(substances, "mass")

    def charge_neutrality_violation(self, substances):
        """Net amount of charge produced

        Parameters
        ----------
        substances: dict

        Returns
        -------
        float: amount of net charge produced/consumed

        """
        return self._violation(substances, "charge")

    def composition_violation(self, substances, composition_keys=None):
        """Net amount of constituent produced

        If composition keys correspond to conserved entities e.g. atoms
        in chemical reactions, this function should return a list of zeros.

        Parameters
        ----------
        substances : dict
        composition_keys : iterable of str, ``None`` or ``True``
            When ``None`` or True: composition keys are taken from substances.
            When ``True`` the keys are also return as an extra return value

        Returns
        -------
        - If ``composition_keys == True``: a tuple: (violations, composition_keys)
        - Otherwise: violations (list of coefficients)

        """
        keys, values = zip(*substances.items())
        ret_comp_keys = composition_keys is True
        if composition_keys in (None, True):
            composition_keys = Substance.composition_keys(values)
        net = [0] * len(composition_keys)
        for substance, coeff in zip(values, self.net_stoich(keys)):
            for idx, key in enumerate(composition_keys):
                net[idx] += substance.composition.get(key, 0) * coeff
        if ret_comp_keys:
            return net, composition_keys
        else:
            return net

    def rate_expr(self):
        """Turns self.param into a RateExpr instance (if not already)

        Default is to create a ``MassAction`` instance. The parameter will
        be used as single instance in ``unique_keys`` if it is a string,
        otherwise it will be used as ``args``.

        Examples
        --------
        >>> r = Reaction.from_string('2 A + B -> 3 C; 7')
        >>> ratex = r.rate_expr()
        >>> ratex.args[0] == 7
        True

        """
        from .util._expr import Expr
        from .kinetics import MassAction

        if isinstance(self.param, Expr):
            return self.param
        else:
            try:
                convertible = self.param.as_RateExpr
            except AttributeError:
                if isinstance(self.param, str):
                    return MassAction.fk(self.param)
                else:
                    return MassAction([self.param])
            else:
                return convertible()

    def rate(self, variables=None, backend=math, substance_keys=None, ratex=None):
        """Evaluate the rate of a reaction

        Parameters
        ----------
        variables : dict
        backend : module, optional
        substance_keys : iterable of str, optional
        ratex : RateExpr

        Returns
        -------
        Dictionary mapping substance keys to the reactions contribution to overall rates.

        Examples
        --------
        >>> rxn1 = Reaction.from_string('2 H2 + O2 -> 2 H2O; 3')
        >>> ref1 = 3*5*5*7
        >>> rxn1.rate({'H2': 5, 'O2': 7}) == {'H2': -2*ref1, 'O2': -ref1, 'H2O': 2*ref1}
        True
        >>> from sympy import Symbol
        >>> k = Symbol('k')
        >>> rxn2 = Reaction(rxn1.reac, rxn1.prod, k)
        >>> concentrations = {key: Symbol(key) for key in set.union(set(rxn1.reac), set(rxn1.prod))}
        >>> import pprint
        >>> pprint.pprint(rxn2.rate(concentrations))
        {'H2': -2*H2**2*O2*k, 'H2O': 2*H2**2*O2*k, 'O2': -H2**2*O2*k}

        """
        if variables is None:
            variables = {}
        if substance_keys is None:
            substance_keys = self.keys()
        if ratex is None:
            ratex = self.rate_expr()

        if isinstance(ratex, Expr):
            srat = ratex(variables, backend=backend, reaction=self)
        else:
            srat = ratex
        return {
            k: srat * v for k, v in zip(substance_keys, self.net_stoich(substance_keys))
        }


def equilibrium_quotient(concs, stoich):
    """Calculates the equilibrium quotient of an equilbrium

    Parameters
    ----------
    concs: array_like
        per substance concentration
    stoich: iterable of integers
        per substance stoichiometric coefficient

    Examples
    --------
    >>> '%.12g' % equilibrium_quotient([1.0, 1e-7, 1e-7], [-1, 1, 1])
    '1e-14'

    """
    import numpy as np

    if not hasattr(concs, "ndim") or concs.ndim == 1:
        tot = 1
    else:
        tot = np.ones(concs.shape[0])
        concs = concs.T

    for nr, conc in zip(stoich, concs):
        tot *= conc ** nr
    return tot


class Equilibrium(Reaction):
    """Represents an equilibrium reaction

    See :class:`Reaction` for parameters

    """

    _str_arrow = "="
    param_char = "K"  # convention

    def check_consistent_units(self, throw=False):
        if is_quantity(self.param):  # This will assume mass action
            exponent = sum(self.prod.values()) - sum(self.reac.values())
            unit_param = unit_of(self.param, simplified=True)
            unit_expected = unit_of(default_units.molar ** exponent, simplified=True)
            if unit_param == unit_expected:
                return True
            else:
                if throw:
                    raise ValueError(
                        "Inconsistent units in equilibrium: %s vs %s"
                        % (unit_param, unit_expected)
                    )
                else:
                    return False
        else:
            return True  # the user might not be using ``chempy.units``

    def as_reactions(
        self,
        kf=None,
        kb=None,
        units=None,
        variables=None,
        backend=math,
        new_name=None,
        **kwargs
    ):
        """Creates a forward and backward :class:`Reaction` pair

        Parameters
        ----------
        kf : float or RateExpr
        kb : float or RateExpr
        units : module
        variables : dict, optional
        backend : module

        """
        nb = sum(self.prod.values())
        nf = sum(self.reac.values())
        if units is None:
            if hasattr(kf, "units") or hasattr(kb, "units"):
                raise ValueError("units missing")
            c0 = 1
        else:
            c0 = 1 * units.molar  # standard concentration IUPAC

        if kf is None:
            fw_name = self.name
            bw_name = new_name
            if kb is None:
                try:
                    kf, kb = self.param
                except TypeError:
                    raise ValueError("Exactly one rate needs to be provided")
            else:
                kf = kb * self.param * c0 ** (nb - nf)
        elif kb is None:
            kb = kf / (self.param * c0 ** (nb - nf))
            fw_name = new_name
            bw_name = self.name
        else:
            raise ValueError("Exactly one rate needs to be provided")

        return (
            Reaction(
                self.reac,
                self.prod,
                kf,
                self.inact_reac,
                self.inact_prod,
                ref=self.ref,
                name=fw_name,
                **kwargs
            ),
            Reaction(
                self.prod,
                self.reac,
                kb,
                self.inact_prod,
                self.inact_reac,
                ref=self.ref,
                name=bw_name,
                **kwargs
            ),
        )

    def equilibrium_expr(self):
        """Turns self.param into a :class:`EqExpr` instance (if not already)

        Examples
        --------
        >>> r = Equilibrium.from_string('2 A + B = 3 C; 7')
        >>> eqex = r.equilibrium_expr()
        >>> eqex.args[0] == 7
        True

        """
        from .util._expr import Expr
        from .thermodynamics import MassActionEq

        if isinstance(self.param, Expr):
            return self.param
        else:
            try:
                convertible = self.param.as_EqExpr
            except AttributeError:
                return MassActionEq([self.param])
            else:
                return convertible()

    def equilibrium_constant(self, variables=None, backend=math):
        """Return equilibrium constant

        Parameters
        ----------
        variables : dict, optional
        backend : module, optional

        """
        return self.equilibrium_expr().eq_const(variables, backend=backend)

    def equilibrium_equation(self, variables, backend=None, **kwargs):
        return self.equilibrium_expr().equilibrium_equation(
            variables, equilibrium=self, backend=backend, **kwargs
        )

    @deprecated(use_instead=equilibrium_constant)
    def K(self, *args, **kwargs):
        return self.equilibrium_constant(*args, **kwargs)

    def Q(self, substances, concs):
        """Calculates the equilibrium qoutient"""
        stoich = self.non_precipitate_stoich(substances)
        return equilibrium_quotient(concs, stoich)

    def precipitate_factor(self, substances, sc_concs):
        factor = 1
        for r, n in self.reac.items():
            if r.precipitate:
                factor *= sc_concs[substances.index(r)] ** -n
        for p, n in self.prod.items():
            if p.precipitate:
                factor *= sc_concs[substances.index(p)] ** n
        return factor

    def dimensionality(self, substances):
        result = 0
        for r, n in self.reac.items():
            if getattr(substances[r], "phase_idx", 0) > 0:
                continue
            result -= n
        for p, n in self.prod.items():
            if getattr(substances[p], "phase_idx", 0) > 0:
                continue
            result += n
        return result

    def __rmul__(self, other):  # This works on both Py2 and Py3
        try:
            other_is_int = other.is_integer
        except AttributeError:
            other_is_int = isinstance(other, int)
        if not other_is_int or not isinstance(self, Equilibrium):
            return NotImplemented
        param = None if self.param is None else self.param ** other
        if other < 0:
            other *= -1
            flip = True
        else:
            flip = False
        other = int(other)  # convert SymPy "Integer" to Python "int"
        reac = dict(other * ArithmeticDict(int, self.reac))
        prod = dict(other * ArithmeticDict(int, self.prod))
        inact_reac = dict(other * ArithmeticDict(int, self.inact_reac))
        inact_prod = dict(other * ArithmeticDict(int, self.inact_prod))
        if flip:
            reac, prod = prod, reac
            inact_reac, inact_prod = inact_prod, inact_reac
        return Equilibrium(
            reac, prod, param, inact_reac=inact_reac, inact_prod=inact_prod
        )

    def __neg__(self):
        return -1 * self

    def __mul__(self, other):
        return other * self

    def __add__(self, other):
        keys = set()
        for key in chain(
            self.reac.keys(), self.prod.keys(), other.reac.keys(), other.prod.keys()
        ):
            keys.add(key)
        reac, prod = {}, {}
        for key in keys:
            n = (
                self.prod.get(key, 0)
                - self.reac.get(key, 0)
                + other.prod.get(key, 0)
                - other.reac.get(key, 0)
            )
            if n < 0:
                reac[key] = -n
            elif n > 0:
                prod[key] = n
            else:
                pass  # n == 0
        if (self.param, other.param) == (None, None):
            param = None
        else:
            param = self.param * other.param
        return Equilibrium(reac, prod, param)

    def __sub__(self, other):
        return self + -1 * other

    @staticmethod
    def eliminate(rxns, wrt):
        """Linear combination coefficients for elimination of a substance

        Parameters
        ----------
        rxns : iterable of Equilibrium instances
        wrt : str (substance key)

        Examples
        --------
        >>> e1 = Equilibrium({'Cd+2': 4, 'H2O': 4}, {'Cd4(OH)4+4': 1, 'H+': 4}, 10**-32.5)
        >>> e2 = Equilibrium({'Cd(OH)2(s)': 1}, {'Cd+2': 1, 'OH-': 2}, 10**-14.4)
        >>> Equilibrium.eliminate([e1, e2], 'Cd+2')
        [1, 4]
        >>> print(1*e1 + 4*e2)
        4 Cd(OH)2(s) + 4 H2O = Cd4(OH)4+4 + 4 H+ + 8 OH-; 7.94e-91

        """
        import sympy

        viol = [r.net_stoich([wrt])[0] for r in rxns]
        factors = defaultdict(int)
        for v in viol:
            for f in sympy.primefactors(v):
                factors[f] = max(factors[f], sympy.Abs(v // f))
        rcd = reduce(mul, (k ** v for k, v in factors.items()))
        viol[0] *= -1
        return [rcd // v for v in viol]

    def cancel(self, rxn):
        """Multiplier of how many times rxn can be added/subtracted.

        Parameters
        ----------
        rxn : Equilibrium

        Examples
        --------
        >>> e1 = Equilibrium({'Cd(OH)2(s)': 4, 'H2O': 4},
        ...                  {'Cd4(OH)4+4': 1, 'H+': 4, 'OH-': 8}, 7.94e-91)
        >>> e2 = Equilibrium({'H2O': 1}, {'H+': 1, 'OH-': 1}, 10**-14)
        >>> e1.cancel(e2)
        -4
        >>> print(e1 - 4*e2)
        4 Cd(OH)2(s) = Cd4(OH)4+4 + 4 OH-; 7.94e-35

        """
        keys = rxn.keys()
        s1 = self.net_stoich(keys)
        s2 = rxn.net_stoich(keys)
        candidate = float("inf")
        for v1, v2 in zip(s1, s2):
            r = intdiv(-v1, v2)
            candidate = min(candidate, r, key=abs)
        return candidate


def _solve_balancing_ilp_pulp(A):
    import pulp

    x = [
        pulp.LpVariable("x%d" % i, lowBound=1, cat="Integer") for i in range(A.shape[1])
    ]
    prob = pulp.LpProblem("chempy_balancing_problem", pulp.LpMinimize)
    prob += reduce(add, x)
    for expr in [
        pulp.lpSum([x[i] * e for i, e in enumerate(row)]) for row in A.tolist()
    ]:
        prob += expr == 0
    prob.solve(pulp.PULP_CBC_CMD(msg=False))
    return [pulp.value(_) for _ in x]


def balance_stoichiometry(
    reactants,
    products,
    substances=None,
    substance_factory=Substance.from_formula,
    parametric_symbols=None,
    underdetermined=True,
    allow_duplicates=False,
):
    """Balances stoichiometric coefficients of a reaction

    Parameters
    ----------
    reactants : iterable of reactant keys
    products : iterable of product keys
    substances : OrderedDict or string or None
        Mapping reactant/product keys to instances of :class:`Substance`.
    substance_factory : callback
    parametric_symbols : generator of symbols
        Used to generate symbols for parametric solution for
        under-determined system of equations. Default is numbered "x-symbols" starting
        from 1.
    underdetermined : bool
        Allows to find a non-unique solution (in addition to a constant factor
        across all terms). Set to ``False`` to disallow (raise ValueError) on
        e.g. "C + O2 -> CO + CO2". Set to ``None`` if you want the symbols replaced
        so that the coefficients are the smallest possible positive (non-zero) integers.
    allow_duplicates : bool
        If False: raises an exception if keys appear in both ``reactants`` and ``products``.

    Examples
    --------
    >>> ref = {'C2H2': 2, 'O2': 3}, {'CO': 4, 'H2O': 2}
    >>> balance_stoichiometry({'C2H2', 'O2'}, {'CO', 'H2O'}) == ref
    True
    >>> ref2 = {'H2': 1, 'O2': 1}, {'H2O2': 1}
    >>> balance_stoichiometry('H2 O2'.split(), ['H2O2'], 'H2 O2 H2O2') == ref2
    True
    >>> reac, prod = 'CuSCN KIO3 HCl'.split(), 'CuSO4 KCl HCN ICl H2O'.split()
    >>> Reaction(*balance_stoichiometry(reac, prod)).string()
    '4 CuSCN + 7 KIO3 + 14 HCl -> 4 CuSO4 + 7 KCl + 4 HCN + 7 ICl + 5 H2O'
    >>> balance_stoichiometry({'Fe', 'O2'}, {'FeO', 'Fe2O3'}, underdetermined=False)
    Traceback (most recent call last):
        ...
    ValueError: The system was under-determined
    >>> r, p = balance_stoichiometry({'Fe', 'O2'}, {'FeO', 'Fe2O3'})
    >>> list(set.union(*[v.free_symbols for v in r.values()]))
    [x1]
    >>> b = balance_stoichiometry({'Fe', 'O2'}, {'FeO', 'Fe2O3'}, underdetermined=None)
    >>> b == ({'Fe': 3, 'O2': 2}, {'FeO': 1, 'Fe2O3': 1})
    True
    >>> d = balance_stoichiometry({'C', 'CO'}, {'C', 'CO', 'CO2'}, underdetermined=None, allow_duplicates=True)
    >>> d == ({'CO': 2}, {'C': 1, 'CO2': 1})
    True

    Returns
    -------
    balanced reactants : dict
    balanced products : dict

    """
    import sympy
    from sympy import (
        MutableDenseMatrix,
        gcd,
        zeros,
        linsolve,
        numbered_symbols,
        nsimplify,
        Wild,
        Symbol,
        Integer,
        Tuple,
        preorder_traversal as pre,
    )

    _intersect = sorted(set.intersection(*map(set, (reactants, products))))
    if _intersect:
        if allow_duplicates:
            if underdetermined is not None:
                raise NotImplementedError(
                    "allow_duplicates currently requires underdetermined=None"
                )
            if set(reactants) == set(products):
                raise ValueError("cannot balance: reactants and products identical")

            # For each duplicate, try to drop it completely:
            for dupl in _intersect:
                try:
                    result = balance_stoichiometry(
                        [sp for sp in reactants if sp != dupl],
                        [sp for sp in products if sp != dupl],
                        substances=substances,
                        substance_factory=substance_factory,
                        underdetermined=underdetermined,
                        allow_duplicates=True,
                    )
                except Exception:
                    continue
                else:
                    return result
            for perm in product(
                *[(False, True)] * len(_intersect)
            ):  # brute force (naive)
                r = set(reactants)
                p = set(products)
                for remove_reac, dupl in zip(perm, _intersect):
                    if remove_reac:
                        r.remove(dupl)
                    else:
                        p.remove(dupl)
                try:
                    result = balance_stoichiometry(
                        r,
                        p,
                        substances=substances,
                        substance_factory=substance_factory,
                        parametric_symbols=parametric_symbols,
                        underdetermined=underdetermined,
                        allow_duplicates=False,
                    )
                except ValueError:
                    continue
                else:
                    return result
            else:
                raise ValueError("Failed to remove duplicate keys: %s" % _intersect)
        else:
            raise ValueError("Substances on both sides: %s" % str(_intersect))
    if substances is None:
        substances = OrderedDict(
            [(k, substance_factory(k)) for k in chain(reactants, products)]
        )
    if isinstance(substances, str):
        substances = OrderedDict(
            [(k, substance_factory(k)) for k in substances.split()]
        )
    if type(reactants) == set:  # noqa
        # we don't want isinstance since it might be "OrderedSet"
        reactants = sorted(reactants)
    if type(products) == set:  # noqa
        products = sorted(products)
    subst_keys = list(reactants) + list(products)

    cks = Substance.composition_keys(substances.values())

    if parametric_symbols is None:
        parametric_symbols = numbered_symbols("x", start=1, integer=True, positive=True)

    # ?C2H2 + ?O2 -> ?CO + ?H2O
    # Ax = 0
    #   A:                    x:
    #
    #   C2H2   O2  CO  H2O
    # C -2     0    1   0      x0    =   0
    # H -2     0    0   2      x1        0
    # O  0    -2    1   1      x2        0
    #                          x3

    def _get(ck, sk):
        return substances[sk].composition.get(ck, 0) * (-1 if sk in reactants else 1)

    for ck in cks:  # check that all components are present on reactant & product sides
        for rk in reactants:
            if substances[rk].composition.get(ck, 0) != 0:
                break
        else:
            any_pos = any(substances[pk].composition.get(ck, 0) > 0 for pk in products)
            any_neg = any(substances[pk].composition.get(ck, 0) < 0 for pk in products)
            if any_pos and any_neg:
                pass  # negative and positive parts among products, no worries
            else:
                raise ValueError("Component '%s' not among reactants" % ck)

        for pk in products:
            if substances[pk].composition.get(ck, 0) != 0:
                break
        else:
            any_pos = any(substances[pk].composition.get(ck, 0) > 0 for pk in reactants)
            any_neg = any(substances[pk].composition.get(ck, 0) < 0 for pk in reactants)
            if any_pos and any_neg:
                pass  # negative and positive parts among reactants, no worries
            else:
                raise ValueError("Component '%s' not among products" % ck)

    A = MutableDenseMatrix([[_get(ck, sk) for sk in subst_keys] for ck in cks])
    A = nsimplify(A)
    symbs = list(reversed([next(parametric_symbols) for _ in range(len(subst_keys))]))
    (sol,) = linsolve((A, zeros(len(cks), 1)), symbs)
    try:
        sol = nsimplify(sol)
    except (AttributeError):
        pass

    wi = Wild("wi", properties=[lambda k: not k.has(Symbol)])
    cd = reduce(
        gcd,
        [1]
        + [
            1 / m[wi]
            for m in map(lambda n: n.match(symbs[-1] / wi), pre(sol))
            if m is not None
        ],
    )
    sol = sol.func(*[arg / cd for arg in sol.args])

    def remove(cont, symb, remaining):
        subsd = dict(zip(remaining / symb, remaining))
        cont = cont.func(*[(arg / symb).expand().subs(subsd) for arg in cont.args])
        if cont.has(symb):
            raise ValueError(
                "Bug, please report an issue at https://github.com/bjodah/chempy"
            )
        return cont

    done = False
    for idx, symb in enumerate(symbs):
        for expr in sol:
            iterable = expr.args if expr.is_Add else [expr]
            for term in iterable:
                if term.is_number:
                    done = True
                    break
            if done:
                break
        if done:
            break
        for expr in sol:
            if (expr / symb).is_number:
                sol = remove(sol, symb, MutableDenseMatrix(symbs[idx + 1 :]))
                break
    for symb in symbs:
        cd = 1
        for expr in sol:
            iterable = expr.args if expr.is_Add else [expr]
            for term in iterable:
                if term.is_Mul and term.args[0].is_number and term.args[1] == symb:
                    cd = gcd(cd, term.args[0])
        if cd != 1:
            sol = sol.func(*[arg.subs(symb, symb / cd) for arg in sol.args])
    integer_one = 1  # need 'is' check, SyntaxWarning when checking against literal
    if underdetermined is integer_one:
        from ._release import __version__

        if int(__version__.split(".")[1]) > 6:
            warnings.warn(  # deprecated because comparison with ``1`` problematic (True==1)
                (
                    "Pass underdetermined == None instead of ``1`` (deprecated since 0.7.0,"
                    " will_be_missing_in='0.9.0')"
                ),
                ChemPyDeprecationWarning,
            )
        underdetermined = None
    if underdetermined is None:
        sol = Tuple(*[Integer(x) for x in _solve_balancing_ilp_pulp(A)])

    fact = gcd(sol)
    sol = MutableDenseMatrix([e / fact for e in sol]).reshape(len(sol), 1)
    sol /= reduce(gcd, sol)
    sol = nsimplify(sol)

    if 0 in sol:
        raise ValueError("Superfluous species given.")
    if underdetermined:
        if any(x == sympy.nan for x in sol):
            raise ValueError("Failed to balance reaction")
    else:
        for x in sol:
            if len(x.free_symbols) != 0:
                raise ValueError("The system was under-determined")
        if not all(residual == 0 for residual in A * sol):
            raise ValueError("Failed to balance reaction")

    def _x(k):
        coeff = sol[subst_keys.index(k)]
        return int(coeff) if underdetermined is None else coeff

    return (
        OrderedDict([(k, _x(k)) for k in reactants]),
        OrderedDict([(k, _x(k)) for k in products]),
    )


def mass_fractions(
    stoichiometries, substances=None, substance_factory=Substance.from_formula
):
    """Calculates weight fractions of each substance in a stoichiometric dict

    Parameters
    ----------
    stoichiometries : dict or set
        If a ``set``: all entries are assumed to correspond to unit multiplicity.
    substances: dict or None

    Examples
    --------
    >>> r = mass_fractions({'H2': 1, 'O2': 1})
    >>> mH2, mO2 = 1.008*2, 15.999*2
    >>> abs(r['H2'] - mH2/(mH2+mO2)) < 1e-4
    True
    >>> abs(r['O2'] - mO2/(mH2+mO2)) < 1e-4
    True
    >>> mass_fractions({'H2O2'}) == {'H2O2': 1.0}
    True

    """
    if isinstance(stoichiometries, set):
        stoichiometries = {k: 1 for k in stoichiometries}
    if substances is None:
        substances = OrderedDict([(k, substance_factory(k)) for k in stoichiometries])
    tot_mass = sum([substances[k].mass * v for k, v in stoichiometries.items()])
    return {k: substances[k].mass * v / tot_mass for k, v in stoichiometries.items()}



# -*- coding: utf-8 -*-

from itertools import chain
from .periodic import groups, symbols, names

_anions = {
    "F-": "fluoride",
    "Cl-": "chloride",
    "Br-": "bromide",
    "I-": "iodide",
    "OH-": "hydroxide",
    "CN-": "cyanide",
    "SCN-": "thiocyanate",
    "CO3-2": "carbonate",
    "C2O4-2": "oxalate",
    "HCO3-": "hydrogencarbonate",
    "NO3-": "nitrate",
    "NO2-": "nitrite",
    "PO4-3": "phospahte",
    "HPO4-2": "hydrogenphospahte",
    "H2PO4-": "dihydrogenphospahte",
    "P-3": "phosphide",
    "SO4-2": "sulphate",
    "HSO4-": "hydrogensulphate",
    "SO3-2": "sulphite",
    "HSO3-": "hydrogensulphite",
    "S-2": "sulfide",
    "ClO-": "hypochlorite",
    "ClO2-": "chlorite",
    "ClO3-": "chlorate",
    "ClO4-": "perchlorate",
    "CrO4-2": "chromate(VI)",
    "Cr2O7-2": "dichromate(VI)",
    "MnO4-2": "manganate(VI)",
    "MnO4-": "permanganate(VII)",
    "FeO4-2": "ferrate(VI)",
    "OsO4-2": "osmate(VI)",
    "Bo3-3": "borate",
    "BiO3-": "bismuthate(V)",
}
_cations = {
    "H3O+": "hydronium",
}
_cation_oxidation_states = {  # This needs to be reviewed, just from the top of my head
    "Cr": (2, 3),
    "Fe": (2, 3),
    "Mn": (2,),
    "Co": (2, 3),
    "Ni": (2, 3),
    "Cu": (1, 2, 3),
    "Ag": (1, 2),
    "Au": (3,),
    "Zn": (2,),
    "Cd": (2,),
    "Hg": (1, 2),  # Tricky: Hg2+2
    "Al": (3,),
    "Ga": (3,),
    "In": (3,),
    "Tl": (1, 3),
    "Sn": (2, 4),
    "Pb": (2, 4),
    "Bi": (3,),
    "Sb": (3,),
}

_alkali = [(symbols[n] + "+", names[n].lower()) for n in groups[1]]
_alkaline_earth = [(symbols[n] + "+2", names[n].lower()) for n in groups[2]]
_all_names = dict(chain(_alkali, _alkaline_earth, _anions.items()))


def name(ion):
    return _all_names[ion]


def ions_from_formula(formula):
    """
    This will be working examples eventually:

    #>>> ions_from_formula('NaCl') == {'Na+': 1, 'Cl-': 1}
    #True
    #>>> ions_from_formula('Fe(NO3)3') == {'Fe+3': 1, 'NO3-': 3}
    #True
    #>>> ions_from_formula('FeSO4') == {'Fe+2': 1, 'SO4-2': 1}
    #True
    #>>> ions_from_formula('(NH4)3PO4') == {'NH4+': 3, 'PO4-3': 1}
    #True
    #>>> ions_from_formula('KAl(SO4)2.11H2O') == {'K+': 1, 'Al+3': 1, 'SO4-2': 2}
    #True

    """
    pass



# -*- coding: utf-8 -*-
"""
This module contains functions for formulating systems of Ordinary Differential
Equations (ODE-systems) which may be integrated numerically to model temporal
evolution of concentrations in reaction systems.
"""

from collections import OrderedDict
from functools import reduce, partial
from itertools import chain
from operator import attrgetter, mul
import math
import warnings

try:
    import numpy as np
except ImportError:
    np = None

from ..units import (
    to_unitless,
    get_derived_unit,
    rescale,
    magnitude,
    unitless_in_registry,
    default_unit_in_registry,
    default_units as u,
)
from ..util.pyutil import deprecated
from ..util._expr import Expr, Symbol
from .rates import RateExpr, MassAction


def _get_derived_unit(reg, key):
    try:
        return get_derived_unit(reg, key)
    except KeyError:
        return get_derived_unit(reg, "_".join(key.split("_")[:-1]))


def law_of_mass_action_rates(conc, rsys, variables=None):
    """Returns a generator of reaction rate expressions

    Rates from the law of mass action (:attr:`Reaction.inact_reac` ignored)
    from a :class:`ReactionSystem`.

    Parameters
    ----------
    conc : array_like
        concentrations (floats or symbolic objects)
    rsys : ReactionSystem instance
        See :class:`ReactionSystem`
    variables : dict (optional)
        to override parameters in the rate expressions of the reactions

    Examples
    --------
    >>> from chempy import ReactionSystem, Reaction
    >>> line, keys = 'H2O -> H+ + OH- ; 1e-4', 'H2O H+ OH-'
    >>> rxn = Reaction.from_string(line, keys)
    >>> rsys = ReactionSystem([rxn], keys)
    >>> next(law_of_mass_action_rates([55.4, 1e-7, 1e-7], rsys))
    0.00554
    >>> from chempy.kinetics.rates import Arrhenius, MassAction
    >>> rxn.param = MassAction(Arrhenius({'A': 1e10, 'Ea_over_R': 9314}))
    >>> print('%.5g' % next(law_of_mass_action_rates([55.4, 1e-7, 1e-7], rsys, {'temperature': 293})))
    0.0086693

    """
    for idx_r, rxn in enumerate(rsys.rxns):
        if isinstance(rxn.param, RateExpr):
            if isinstance(rxn.param, MassAction):
                yield rxn.param(
                    dict(chain(variables.items(), zip(rsys.substances.keys(), conc))),
                    reaction=rxn,
                )
            else:
                raise ValueError("Not mass-action rate in reaction %d" % idx_r)
        else:
            rate = 1
            for substance_key, coeff in rxn.reac.items():
                s_idx = rsys.as_substance_index(substance_key)
                rate *= conc[s_idx] ** coeff
            yield rate * rxn.param


def dCdt_list(rsys, rates):
    """Returns a list of the time derivatives of the concentrations

    Parameters
    ----------
    rsys: ReactionSystem instance
    rates: array_like
        rates (to be weighted by stoichiometries) of the reactions
        in ``rsys``

    Examples
    --------
    >>> from chempy import ReactionSystem, Reaction
    >>> line, keys = 'H2O -> H+ + OH- ; 1e-4', 'H2O H+ OH-'
    >>> rsys = ReactionSystem([Reaction.from_string(line, keys)], keys)
    >>> dCdt_list(rsys, [0.0054])
    [-0.0054, 0.0054, 0.0054]

    """
    f = [0] * rsys.ns
    net_stoichs = rsys.net_stoichs()
    for idx_s in range(rsys.ns):
        for idx_r in range(rsys.nr):
            f[idx_s] += net_stoichs[idx_r, idx_s] * rates[idx_r]
    return f


def get_odesys(
    rsys,
    include_params=True,
    substitutions=None,
    SymbolicSys=None,
    unit_registry=None,
    output_conc_unit=None,
    output_time_unit=None,
    cstr=False,
    constants=None,
    **kwargs
):
    """Creates a :class:`pyneqsys.SymbolicSys` from a :class:`ReactionSystem`

    The parameters passed to RateExpr will contain the key ``'time'`` corresponding to the
    independent variable of the IVP.

    Parameters
    ----------
    rsys : ReactionSystem
        Each reaction of ``rsys`` will have their :meth:`Reaction.rate_expr()` invoked.
        Note that if :attr:`Reaction.param` is not a :class:`RateExpr` (or convertible to
        one through :meth:`as_RateExpr`) it will be used to construct a :class:`MassAction`
        instance.
    include_params : bool (default: True)
        Whether rate constants should be included into the rate expressions or
        left as free parameters in the :class:`pyneqsys.SymbolicSys` instance.
    substitutions : dict, optional
        Variable substitutions used by rate expressions (in respective Reaction.param).
        values are allowed to be values of instances of :class:`Expr`.
    SymbolicSys : class (optional)
        Default : :class:`pyneqsys.SymbolicSys`.
    unit_registry: dict (optional)
        See :func:`chempy.units.get_derived_units`.
    output_conc_unit : unit (Optional)
    output_time_unit : unit (Optional)
    cstr : bool
        Generate expressions for continuously stirred tank reactor.
    constants : module
        e.g. ``chempy.units.default_constants``, parameter keys not found in
        substitutions will be looked for as an attribute of ``constants`` when provided.
    \\*\\*kwargs :
        Keyword arguments passed on to `SymbolicSys`.

    Returns
    -------
    pyodesys.symbolic.SymbolicSys
    extra : dict, with keys:
        - param_keys : list of str instances
        - unique : OrderedDict mapping str to value (possibly None)
        - p_units : list of units
        - max_euler_step_cb : callable or None
        - linear_dependencies : None or factory of solver callback
        - rate_exprs_cb : callable
        - cstr_fr_fc : None or (feed-ratio-key, subtance-key-to-feed-conc-key-map)

    Examples
    --------
    >>> from chempy import Equilibrium, ReactionSystem
    >>> eq = Equilibrium({'Fe+3', 'SCN-'}, {'FeSCN+2'}, 10**2)
    >>> substances = 'Fe+3 SCN- FeSCN+2'.split()
    >>> rsys = ReactionSystem(eq.as_reactions(kf=3.0), substances)
    >>> odesys, extra = get_odesys(rsys)
    >>> init_conc = {'Fe+3': 1.0, 'SCN-': .3, 'FeSCN+2': 0}
    >>> tout, Cout, info = odesys.integrate(5, init_conc)
    >>> Cout[-1, :].round(4)
    array([0.7042, 0.0042, 0.2958])

    """
    if SymbolicSys is None:
        from pyodesys.symbolic import SymbolicSys

    r_exprs = [rxn.rate_expr() for rxn in rsys.rxns]
    _ori_pk = set.union(*(ratex.all_parameter_keys() for ratex in r_exprs))
    _ori_uk = set.union(*(ratex.all_unique_keys() for ratex in r_exprs))
    _subst_pk = set()
    _active_subst = OrderedDict()
    _passive_subst = OrderedDict()
    substitutions = substitutions or {}

    unique = OrderedDict()
    unique_units = {}

    cstr_fr_fc = (
        ("feedratio", OrderedDict([(sk, "fc_" + sk) for sk in rsys.substances]))
        if cstr is True
        else cstr
    )

    if cstr_fr_fc:
        _ori_pk.add(cstr_fr_fc[0])
        for k in cstr_fr_fc[1].values():
            _ori_pk.add(k)

    def _reg_unique_unit(k, arg_dim, idx):
        if unit_registry is None:
            return
        unique_units[k] = reduce(
            mul, [1] + [unit_registry[dim] ** v for dim, v in arg_dim[idx].items()]
        )

    def _get_arg_dim(expr, rxn):
        if unit_registry is None:
            return None
        else:
            return expr.args_dimensionality(reaction=rxn)

    def _reg_unique(expr, rxn=None):
        if not isinstance(expr, Expr):
            raise NotImplementedError("Currently only Expr sub classes are supported.")

        if isinstance(expr, MassAction):
            if expr.args is None:
                (uk,) = expr.unique_keys
                if uk not in substitutions:
                    unique[uk] = None
                    _reg_unique_unit(uk, _get_arg_dim(expr, rxn), 0)
                    return
            else:
                (arg,) = expr.args
                if isinstance(arg, Symbol):
                    (uk,) = arg.unique_keys
                    if uk not in substitutions:
                        unique[uk] = None
                        _reg_unique_unit(uk, _get_arg_dim(expr, rxn), 0)
                        return

        if expr.args is None:
            for idx, k in enumerate(expr.unique_keys):
                if k not in substitutions:
                    unique[k] = None
                    _reg_unique_unit(k, _get_arg_dim(expr, rxn), idx)
        else:
            for idx, arg in enumerate(expr.args):
                if isinstance(arg, Expr):
                    _reg_unique(arg, rxn=rxn)
                elif expr.unique_keys is not None and idx < len(expr.unique_keys):
                    uk = expr.unique_keys[idx]
                    if uk not in substitutions:
                        unique[uk] = arg
                        _reg_unique_unit(uk, _get_arg_dim(expr, rxn), idx)

    for sk, sv in substitutions.items():
        if sk not in _ori_pk and sk not in _ori_uk:
            raise ValueError(
                "Substitution: '%s' does not appear in any rate expressions." % sk
            )
        if isinstance(sv, Expr):
            _subst_pk.update(sv.parameter_keys)
            _active_subst[sk] = sv
            if not include_params:
                _reg_unique(sv)
        else:
            # if unit_registry is None:
            if unit_registry is not None:
                sv = unitless_in_registry(sv, unit_registry)
            _passive_subst[sk] = sv

    all_pk = []
    for pk in filter(
        lambda x: x not in substitutions and x != "time", _ori_pk.union(_subst_pk)
    ):
        if hasattr(constants, pk):
            const = getattr(constants, pk)
            if unit_registry is None:
                const = magnitude(const)
            else:
                const = unitless_in_registry(const, unit_registry)

            _passive_subst[pk] = const
        else:
            all_pk.append(pk)

    if not include_params:
        for rxn, ratex in zip(rsys.rxns, r_exprs):
            _reg_unique(ratex, rxn)

    all_pk_with_unique = list(
        chain(all_pk, filter(lambda k: k not in all_pk, unique.keys()))
    )
    if include_params:
        param_names_for_odesys = all_pk
    else:
        param_names_for_odesys = all_pk_with_unique

    if unit_registry is None:
        p_units = None
    else:
        # We need to make rsys_params unitless and create
        # a pre- & post-processor for SymbolicSys
        pk_units = [_get_derived_unit(unit_registry, k) for k in all_pk]
        p_units = (
            pk_units
            if include_params
            else (pk_units + [unique_units[k] for k in unique])
        )
        new_r_exprs = []
        for ratex in r_exprs:
            _pu, _new_ratex = ratex.dedimensionalisation(unit_registry)
            new_r_exprs.append(_new_ratex)
        r_exprs = new_r_exprs

        time_unit = get_derived_unit(unit_registry, "time")
        conc_unit = get_derived_unit(unit_registry, "concentration")

        def post_processor(x, y, p):
            time = x * time_unit
            if output_time_unit is not None:
                time = rescale(time, output_time_unit)
            conc = y * conc_unit
            if output_conc_unit is not None:
                conc = rescale(conc, output_conc_unit)
            return (
                time,
                conc,
                np.array(
                    [elem * p_unit for elem, p_unit in zip(p.T, p_units)], dtype=object
                ).T,
            )

        kwargs["to_arrays_callbacks"] = (
            lambda x: to_unitless(x, time_unit),
            lambda y: to_unitless(y, conc_unit),
            lambda p: np.array(
                [
                    to_unitless(px, p_unit)
                    for px, p_unit in zip(p.T if hasattr(p, "T") else p, p_units)
                ]
            ).T,
        )
        kwargs["post_processors"] = kwargs.get("post_processors", []) + [post_processor]

    def dydt(t, y, p, backend=math):
        variables = dict(chain(y.items(), p.items()))
        if "time" in variables:
            raise ValueError("Key 'time' is reserved.")
        variables["time"] = t
        for k, act in _active_subst.items():
            if unit_registry is not None and act.args:
                _, act = act.dedimensionalisation(unit_registry)
            variables[k] = act(variables, backend=backend)
        variables.update(_passive_subst)
        return rsys.rates(
            variables, backend=backend, ratexs=r_exprs, cstr_fr_fc=cstr_fr_fc
        )

    def reaction_rates(t, y, p, backend=math):
        variables = dict(chain(y.items(), p.items()))
        if "time" in variables:
            raise ValueError("Key 'time' is reserved.")
        variables["time"] = t
        for k, act in _active_subst.items():
            if unit_registry is not None and act.args:
                _, act = act.dedimensionalisation(unit_registry)
            variables[k] = act(variables, backend=backend)
        variables.update(_passive_subst)
        return [
            ratex(variables, backend=backend, reaction=rxn)
            for rxn, ratex in zip(rsys.rxns, r_exprs)
        ]

    names = [s.name for s in rsys.substances.values()]
    latex_names = [
        None if s.latex_name is None else ("\\mathrm{" + s.latex_name + "}")
        for s in rsys.substances.values()
    ]

    compo_vecs, compo_names = rsys.composition_balance_vectors()

    odesys = SymbolicSys.from_callback(
        dydt,
        dep_by_name=True,
        par_by_name=True,
        names=names,
        latex_names=latex_names,
        param_names=param_names_for_odesys,
        linear_invariants=None if len(compo_vecs) == 0 else compo_vecs,
        linear_invariant_names=None
        if len(compo_names) == 0
        else list(map(str, compo_names)),
        **kwargs
    )

    symbolic_ratexs = reaction_rates(
        odesys.indep,
        dict(zip(odesys.names, odesys.dep)),
        dict(zip(odesys.param_names, odesys.params)),
        backend=odesys.be,
    )
    rate_exprs_cb = odesys._callback_factory(symbolic_ratexs)

    if rsys.check_balance(strict=True):
        # Composition available, we can provide callback for calculating
        # maximum allowed Euler forward step at start of integration.
        def max_euler_step_cb(x, y, p=()):
            _x, _y, _p = odesys.pre_process(*odesys.to_arrays(x, y, p))
            upper_bounds = rsys.upper_conc_bounds(_y)
            fvec = odesys.f_cb(_x[0], _y, _p)
            h = []
            for idx, fcomp in enumerate(fvec):
                if fcomp == 0:
                    h.append(float("inf"))
                elif fcomp > 0:
                    h.append((upper_bounds[idx] - _y[idx]) / fcomp)
                else:  # fcomp < 0
                    h.append(-_y[idx] / fcomp)
            min_h = min(h)
            return min(min_h, 1)

        def linear_dependencies(preferred=None):
            if preferred is not None:
                if len(preferred) == 0:
                    raise ValueError("No preferred substance keys provided")
                if len(preferred) >= len(rsys.substances):
                    raise ValueError(
                        "Cannot remove all concentrations from linear dependencies"
                    )
                for k in preferred:
                    if k not in rsys.substances:
                        raise ValueError("Unknown substance key: %s" % k)

            def analytic_solver(x0, y0, p0, be):
                if preferred is None:
                    _preferred = None
                else:
                    _preferred = list(preferred)
                A = be.Matrix(compo_vecs)
                rA, pivots = A.rref()

                analytic_exprs = OrderedDict()
                for ri, ci1st in enumerate(pivots):
                    for idx in range(ci1st, odesys.ny):
                        key = odesys.names[idx]
                        if rA[ri, idx] == 0:
                            continue
                        if _preferred is None or key in _preferred:
                            terms = [
                                rA[ri, di] * (odesys.dep[di] - y0[odesys.dep[di]])
                                for di in range(ci1st, odesys.ny)
                                if di != idx
                            ]
                            analytic_exprs[odesys[key]] = (
                                y0[odesys.dep[idx]] - sum(terms) / rA[ri, idx]
                            )
                            if _preferred is not None:
                                _preferred.remove(key)
                            break
                for k in reversed(list(analytic_exprs.keys())):
                    analytic_exprs[k] = analytic_exprs[k].subs(analytic_exprs)
                if _preferred is not None and len(_preferred) > 0:
                    raise ValueError(
                        "Failed to obtain analytic expression for: %s"
                        % ", ".join(_preferred)
                    )
                return analytic_exprs

            return analytic_solver

    else:
        max_euler_step_cb = None
        linear_dependencies = None

    return odesys, {
        "param_keys": all_pk,
        "unique": unique,
        "p_units": p_units,
        "max_euler_step_cb": max_euler_step_cb,
        "linear_dependencies": linear_dependencies,
        "rate_exprs_cb": rate_exprs_cb,
        "cstr_fr_fc": cstr_fr_fc,
        "unit_registry": unit_registry,
    }


@deprecated(
    last_supported_version="0.5.3",
    will_be_missing_in="0.8.0",
    use_instead="pyodesys.chained_parameter_variation",
)
def chained_parameter_variation(
    odesys, durations, init_conc, varied_params, default_params, integrate_kwargs=None
):
    """Integrate an ODE-system for a serie of durations with some parameters changed in-between

    Parameters
    ----------
    odesys : :class:`pyodesys.ODESys` instance
    durations : iterable of floats
    init_conc : dict or array_like
    varied_params : dict mapping parameter name to array_like
        Each array_like need to be of same length as durations.
    default_params : dict or array_like
        Default values for the parameters of the ODE system.
    integrate_kwargs : dict
        Keyword arguments passed on to :meth:`pyodesys.ODESys.integrate`.

    """
    for k, v in varied_params.items():
        if len(v) != len(durations):
            raise ValueError("Mismathced lengths of durations and varied_params")
    integrate_kwargs = integrate_kwargs or {}
    touts = []
    couts = []
    infos = {}
    c0 = init_conc.copy()
    for idx, duration in enumerate(durations):
        params = default_params.copy()
        for k, v in varied_params.items():
            params[k] = v[idx]
        tout, cout, info = odesys.integrate(duration, c0, params, **integrate_kwargs)
        c0 = cout[-1, :]
        idx0 = 0 if idx == 0 else 1
        t_global = 0 if idx == 0 else touts[-1][-1]
        touts.append(tout[idx0:] + t_global)
        couts.append(cout[idx0:, ...])
        for k, v in info.items():
            if k.startswith("internal"):
                continue
            if k in infos:
                infos[k] += (v,)
            else:
                infos[k] = (v,)
    return np.concatenate(touts), np.concatenate(couts), infos


def _create_odesys(
    rsys,
    substance_symbols=None,
    parameter_symbols=None,
    pretty_replace=lambda x: x,
    backend=None,
    SymbolicSys=None,
    time_symbol=None,
    unit_registry=None,
    rates_kw=None,
    parameter_expressions=None,
    symbolic_kw=None,
):
    """This will be a simpler version of get_odesys without the unit handling code.
    The motivation is to reduce complexity (the code of get_odesys is long with multiple closures).

    This will also rely on SymPy explicitly and the user will be expected to deal with SymPy
    expressions.

    Only when this function has the same capabilities as get_odesys will it become and public API
    (along with a deprecation of get_odesys).

    Parameters
    ----------
    rsys : ReactionSystem instance
    substance_symbols : OrderedDict
       If ``None``: ``rsys.substances`` will be used.
    parameter_symbols : OrderedDict
    backend : str or module
        Symbolic backend (e.g. sympy). The package ``sym`` is used as a wrapper.
    SymbolicSys: class
        See ``pyodesys`` for API.
    time_symbol : Symbol
    unit_registry : object
        e.g. ``chempy.units.SI_base_registry``
    rates_kw : dict
        Keyword arguments passed to the ``rates`` method of rsys.
    parameter_expressions : dict
        Optional overrides.
    symbolic_kw : dict
        Keyword arguments passed on to SymbolicSys.

    Returns
    -------
    SymbolicSys (subclass of ``pyodesys.ODESys``)
    dict :
        - ``'symbols'``: dict mapping str to symbol.
        - ``'validate'``: callable acppeting a dictionary mapping str to quantities
    """
    if backend is None:
        from sym import Backend

        backend = Backend(backend)
    if SymbolicSys is None:
        from pyodesys.symbolic import SymbolicSys

    if substance_symbols is None:
        substance_symbols = OrderedDict(
            [(key, backend.Symbol(key)) for key in rsys.substances]
        )
    if isinstance(substance_symbols, OrderedDict):
        if list(substance_symbols) != list(rsys.substances):
            raise ValueError(
                "substance_symbols needs to have same (oredered) keys as rsys.substances"
            )

    if parameter_symbols is None:
        keys = []
        for rxnpar in map(attrgetter("param"), rsys.rxns):
            if isinstance(rxnpar, str):
                if rxnpar in (parameter_expressions or {}):
                    for pk in parameter_expressions[rxnpar].all_parameter_keys():
                        keys.append(pk)
                else:
                    keys.append(rxnpar)
            elif isinstance(rxnpar, Expr):
                keys.extend(rxnpar.all_unique_keys())
                for pk in rxnpar.all_parameter_keys():
                    if pk not in keys:
                        keys.append(pk)
            else:
                raise NotImplementedError("Unknown")
        if rates_kw and "cstr_fr_fc" in rates_kw:
            flowrate_volume, feed_conc = rates_kw["cstr_fr_fc"]
            keys.append(flowrate_volume)
            keys.extend(feed_conc.values())
            assert all(sk in rsys.substances for sk in feed_conc)
        if len(keys) != len(set(keys)):
            raise ValueError("Duplicates in keys")
        parameter_symbols = OrderedDict([(key, backend.Symbol(key)) for key in keys])

    if not isinstance(parameter_symbols, OrderedDict):
        raise ValueError("parameter_symbols needs to be an OrderedDict")

    symbols = OrderedDict(chain(substance_symbols.items(), parameter_symbols.items()))
    symbols["time"] = time_symbol or backend.Symbol("t")
    if any(symbols["time"] == v for k, v in symbols.items() if k != "time"):
        raise ValueError("time_symbol already in use (name clash?)")
    varbls = dict(symbols, **parameter_symbols)
    varbls.update(parameter_expressions or {})
    rates = rsys.rates(varbls, **(rates_kw or {}))
    compo_vecs, compo_names = rsys.composition_balance_vectors()

    odesys = SymbolicSys(
        zip(
            [substance_symbols[key] for key in rsys.substances],
            [rates[key] for key in rsys.substances],
        ),
        symbols["time"],
        parameter_symbols.values(),
        names=list(rsys.substances.keys()),
        latex_names=[s.latex_name for s in rsys.substances.values()],
        param_names=parameter_symbols.keys(),
        latex_param_names=[pretty_replace(n) for n in parameter_symbols.keys()],
        linear_invariants=compo_vecs,
        linear_invariant_names=list(map(str, compo_names)),
        backend=backend,
        dep_by_name=True,
        par_by_name=True,
        **(symbolic_kw or {})
    )

    validate = partial(
        _validate, rsys=rsys, symbols=symbols, odesys=odesys, backend=backend
    )
    return odesys, {
        "symbols": symbols,
        "validate": validate,
        "unit_aware_solve": _mk_unit_aware_solve(
            odesys, unit_registry, validate=validate
        )
        if unit_registry
        else None,
    }


def _mk_dedim(unit_registry):
    unit_time = get_derived_unit(unit_registry, "time")
    unit_conc = get_derived_unit(unit_registry, "concentration")

    def dedim_tcp(
        t, c, p, param_unit=lambda k, v: default_unit_in_registry(v, unit_registry)
    ):
        _t = to_unitless(t, unit_time)
        _c = to_unitless(c, unit_conc)
        _p, pu = {}, {}
        for k, v in p.items():
            pu[k] = param_unit(k, v)
            _p[k] = to_unitless(v, pu[k])
        return (_t, _c, _p), dict(
            unit_time=unit_time, unit_conc=unit_conc, param_units=pu
        )

    return locals()


def _mk_unit_aware_solve(odesys, unit_registry, validate):

    dedim_ctx = _mk_dedim(unit_registry)

    def solve(t, c, p, **kwargs):
        for name in odesys.names:
            c[name]  # to e.g. populate defaultdict
        validate(dict(c, **p))
        tcp, dedim_extra = dedim_ctx["dedim_tcp"](t, c, p)
        result = odesys.integrate(*tcp, **kwargs)
        result.xout = result.xout * dedim_extra["unit_time"]
        result.yout = (
            result.yout * dedim_extra["unit_conc"]
        )  # assumes only concentrations in y
        return result, dedim_extra

    return solve


def _exact(v):
    if hasattr(v, "_uncertainty"):
        return v.magnitude * v.units
    else:
        return v


def _validate(
    conditions,
    rsys,
    symbols,
    odesys,
    backend=None,
    transform=None,
    ignore=("time",),
    check_conditions_no_extra=False,
):
    """For use with create_odesys

    Parameters
    ----------
    conditions : OrderedDict
        Parameters, values with units from ``chempy.units``.
    rsys : ReactionSystem
    symbols : dict
        Mapping variable name to symbols.
    backend : module
        Module for symbolic mathematics. (defaults to SymPy)
    transform : callable for rewriting expressions
    check_conditions_no_extra : bool
        When True, conditions may not contain keys not referenced in any expression.

    Raises
    ------
    ``KeyError`` if a key in conditions is not in odesys.names or odesys.param_names

    """
    if backend is None:
        from sym import Backend

        backend = Backend(backend)

    if transform is None:
        if backend.__name__ != "sympy":
            warnings.warn("Backend != SymPy, provide your own transform function.")

        def transform(arg):
            expr = backend.logcombine(arg, force=True)
            v, w = map(backend.Wild, "v w".split())
            return expr.replace(backend.log(w ** v), v * backend.log(w))

    rates = {}
    seen = set()

    for k, v in rsys.rates(symbols).items():
        expr = transform(v)
        if expr == 0:
            rate = 0 * u.molar / u.second
        else:
            rate = None
            for term in (
                expr.args if hasattr(expr, "is_Add") and expr.is_Add else (expr,)
            ):
                args = sorted(expr.free_symbols, key=lambda e: e.name)
                values = [conditions[s.name] for s in args]
                result = backend.lambdify(args, term)(*map(_exact, values))
                to_unitless(
                    result, u.molar / u.second
                )  # raises an exception upon unit error
                if rate is None:
                    rate = result
                else:
                    rate += result
        rates[k] = rate
        seen |= set([s.name for s in expr.free_symbols])
    if check_conditions_no_extra:
        for k in conditions:
            if (
                k not in odesys.param_names
                and k not in odesys.names
                and k not in ignore
            ):
                raise KeyError("Unknown param: %s" % k)
    return {"not_seen": (set(rsys.substances) | set(conditions)) - seen, "rates": rates}



from contextlib import contextmanager
import os
import shutil
import sys
import time
import logging
import inspect
import pprint
import subprocess
import textwrap


class Echo:
    """Context maganger for echoing variable assignments (in CPython)"""

    def __init__(self, msg, indent="  "):
        self.msg = msg
        self.indent = indent
        self.parent_frame = inspect.currentframe().f_back

    def __enter__(self):
        print(self.msg)
        self.locals_on_entry = self.parent_frame.f_locals.copy()

    def __exit__(self, exc_t, exc_v, tb):
        new_locals = dict(
            (k, v)
            for k, v in self.parent_frame.f_locals.items()
            if k not in self.locals_on_entry
        )
        print(textwrap.indent(pprint.pformat(new_locals), self.indent))


class Notify:
    def __init__(self, msg=lambda t: "Job finished in %.1f seconds" % t):
        self.msg = msg
        self.t0 = time.time()

    def __enter__(self):
        pass

    def notify(self, title, message):
        if os.environ.get("DISPLAY", ""):
            subprocess.call(["notify-send", title, message])
        else:
            print(title)
            print(message)

    def __exit__(self, exc_t, exc_v, tb):
        if exc_t is None:
            title = "Success"
        else:
            title = "Failure"
        self.notify(title, self.msg(time.time() - self.t0))


class c:

    ok = "\033[92m"  # green
    fail = "\033[91m"  # red
    endc = "\033[0m"  # reset


class Timed:
    """Utility function for timing portions of your python script

    Parameters
    ----------
    msg : str
    timer : callable
        You can switch to e.g. ``time.process_time`` but note that if other
        programs are called from e.g. ``subprocess.Popen`` the time spent
        in those subprocesses will not be included.

    Examples
    --------
    >>> t = Timed("Counting stars...").tic(); stars.count(); t.toc_and_print()  # doctest: +SKIP
    Counting stars...                                          (42.0 s) [  ok]
    >>> with Timed("Counting sheep..."):  # doctest: +SKIP
    ...     n_sheep = animals.count('sheep')
    ...
    Counting sheep...                                          (17.2 s) [  ok]

    """

    counting = False

    def __init__(self, msg=None, timer=time.time, fmt_s=".1f", out=sys.stdout):
        self.msg = msg
        self.out = out
        self.fmt_s = fmt_s
        sys.stdout.flush()
        self.timer = timer

    def tic(self):
        if self.msg is not None:
            self.out.write(self.msg)
            self.out.flush()
        self.counting = True
        self.t = self.timer()
        return self  # Allows t = Timed().tic(); integrals.calc(); t = t.toc()

    def toc(self, ok=True):
        if self.counting:
            t = self.timer() - self.t
            if self.msg is not None:
                if ok:
                    status = "ok"
                    color = c.ok
                else:
                    status = "error"
                    color = c.fail
                self.out.write(
                    "%{}s\n".format(shutil.get_terminal_size()[0] - len(self.msg))
                    % (
                        "(%{fmt_s} s) [{c}%5s{r}]".format(
                            fmt_s=self.fmt_s, c=color, r=c.endc
                        )
                        % (t, status)
                    )
                )
                self.out.flush()
            return t
        else:
            raise ValueError("Not counting, did you forget to call ``.tic()`` method?")

    def __enter__(self):
        self.tic()

    def __exit__(self, exc_type, exc_value, traceback):
        self.toc(not exc_type)


@contextmanager
def limit_logging(max_lvl=logging.CRITICAL):
    """Contextmanager for silencing logging messages.

    Examples
    --------
    >>> with limit_logging():
    ...     logger.info("you won't see this...")  # doctest: +SKIP

    """
    _ori = logging.root.manager.disable
    logging.disable(max_lvl)
    try:
        yield
    finally:
        logging.disable(_ori)



# -*- coding: utf-8 -*-

import math

from chempy.chemistry import Equilibrium
from chempy.util._expr import Expr
from chempy.util.testing import requires
from chempy.units import (
    allclose,
    units_library,
    Backend,
    default_units as du,
    default_constants as dc,
)


from ..expressions import MassActionEq, GibbsEqConst


@requires("sympy")
def test_MassActionEq_symbolic():
    import sympy as sp

    K, A, B, C = sp.symbols("K A B C")
    mae = MassActionEq([K])
    eq = Equilibrium({"A"}, {"B", "C"})
    expr = mae.equilibrium_equation({"A": A, "B": B, "C": C}, equilibrium=eq)
    assert expr - K + B * C / A == 0


def test_GibbsEqConst():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    gee = GibbsEqConst([dH / R, dS / R])
    ref = math.exp(-(dH - T * dS) / (R * T))
    assert abs((gee({"temperature": T}) - ref) / ref) < 1e-14


def _gibbs(args, T, R, backend):
    H, S, Cp, Tref = args
    H2 = H + Cp * (T - Tref)
    S2 = S + Cp * backend.log(T / Tref)
    return backend.exp(-(H2 - T * S2) / (R * T))


def test_custom_gibbs():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    MyGibbs = MassActionEq.from_callback(
        _gibbs,
        parameter_keys=("temperature", "R"),
        argument_names=("H", "S", "Cp", "Tref"),
    )
    dCp = 123.45
    Tref = 242
    gee2 = MyGibbs([dH, dS, dCp, Tref])
    dH2 = dH + dCp * (T - Tref)
    dS2 = dS + dCp * math.log(T / Tref)
    ref2 = math.exp(-(dH2 - T * dS2) / (R * T))
    assert abs((gee2.eq_const({"temperature": T, "R": R}) - ref2) / ref2) < 1e-14


def test_GibbsEqConst__unique_keys():
    R, T = 8.314, 298.15
    dH, dS = -4e3, 16
    gee = GibbsEqConst(unique_keys=("dH1", "dS1"))
    ref = math.exp(-(dH - T * dS) / (R * T))
    assert (
        abs(
            (gee.eq_const({"temperature": T, "dH1": dH / R, "dS1": dS / R}) - ref) / ref
        )
        < 1e-14
    )


@requires("sympy")
def test_GibbsEqConst__latex():
    import sympy

    DH, DS, R, T = sympy.symbols(r"\Delta\ H \Delta\ S R T")
    gee = GibbsEqConst([DH / R, DS / R])
    res = gee.eq_const({"temperature": T}, backend=sympy)
    ref = sympy.exp(-(DH - T * DS) / (R * T))
    assert (res - ref).simplify() == 0


@requires(units_library)
def test_GibbsEqConst__units():
    R, T = dc.molar_gas_constant, 298.15 * du.K
    DH = -4e3 * du.J / du.mol
    DS = 16 * du.J / du.K / du.mol
    be = Backend()
    gee = GibbsEqConst([DH / R, DS / R])
    res = gee.eq_const({"temperature": T}, backend=be)
    ref = be.exp(-(DH - T * DS) / (R * T))
    assert allclose(res, ref)


@requires(units_library)
def test_GibbsEqConst__nested():
    class TExpr(Expr):
        argument_names = ("heat_capacity",)
        parameter_keys = ("energy",)

        def __call__(self, variables, backend=None):
            (heat_capacity,) = self.all_args(variables, backend=backend)
            (energy,) = self.all_params(variables, backend=backend)
            return energy / heat_capacity

    R = 8.314 * du.J / du.K / du.mol
    T = TExpr([10.0 * du.J / du.K])
    dH, dS = -4e3 * du.J / du.mol, 16 * du.J / du.K / du.mol
    gee = GibbsEqConst([dH / R, dS / R])
    be = Backend()
    Tref = 298.15 * du.K
    ref = be.exp(-(dH - Tref * dS) / (R * Tref))
    assert (
        be.abs(
            (
                gee.eq_const({"energy": 2981.5 * du.J, "temperature": T}, backend=be)
                - ref
            )
            / ref
        )
        < 1e-14
    )



def scrub_output_pre_save(model, **kwargs):
    """scrub output before saving notebooks"""
    # only run on notebooks
    if model["type"] != "notebook":
        return
    # only run on nbformat v4
    if model["content"]["nbformat"] != 4:
        return

    for cell in model["content"]["cells"]:
        if cell["cell_type"] != "code":
            continue
        cell["outputs"] = []
        cell["metadata"] = {}
        cell["execution_count"] = None


c.FileContentsManager.pre_save_hook = scrub_output_pre_save



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from chempy.kinetics.integrated import (
    pseudo_irrev,
    pseudo_rev,
    binary_irrev,
    binary_rev,
)

import sympy

funcs = (pseudo_irrev, pseudo_rev, binary_irrev, binary_rev)


def main():
    """
    This example demonstrates how to generate pretty equations from the analytic
    expressions found in ``chempy.kinetics.integrated``.
    """
    t, kf, t0, major, minor, prod, beta = sympy.symbols(
        "t k_f t0 Y Z X beta", negative=False
    )
    for f in funcs:
        args = [t, kf, prod, major, minor]
        if f in (pseudo_rev, binary_rev):
            args.insert(2, kf / beta)
        expr = f(*args, backend="sympy")
        with open(f.__name__ + ".png", "wb") as ofh:
            sympy.printing.preview(
                expr,
                output="png",
                filename="out.png",
                viewer="BytesIO",
                outputbuffer=ofh,
            )
        with open(f.__name__ + "_diff.png", "wb") as ofh:
            sympy.printing.preview(
                expr.diff(t).subs({t0: 0}).simplify(),
                output="png",
                filename="out.png",
                viewer="BytesIO",
                outputbuffer=ofh,
            )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-


import warnings

from .._util import _any, get_backend


def water_permittivity(
    T=None, P=None, units=None, U=None, just_return_U=False, warn=True, backend=None
):
    """
    Relative permittivity of water as function of temperature (K)
    and pressure (bar).

    Parameters
    ----------
    T : float
        Temperature (default: 298.15 Kelvin)
    P : float
        Pressure (default: 1 bar)
    units : object (optional)
        object with attributes: Kelvin, bar
    U : array_like (optional)
        9 parameters to the equation.
    just_return_U : bool (optional, default: False)
        Do not compute relative permittivity, just return the parameters ``U``.
    warn : bool (default: True)
        Emit UserWarning when outside temperature/pressure range.
    backend : module (default: None)
        modules which contains "exp", default: numpy, math

    Returns
    -------
    Relative permittivity of water (dielectric constant)

    References
    ----------
    Bradley, D.J.; Pitzer, K.S. `Thermodynamics of electrolytes. 12. Dielectric
        properties of water and Debye--Hueckel parameters to 350/sup
        0/C and 1 kbar`, J. Phys. Chem.; Journal Volume 83 (12)
        (1979), pp. 1599-1603,
        http://pubs.acs.org/doi/abs/10.1021/j100475a009
        DOI: 10.1021/j100475a009
    """
    be = get_backend(backend)
    if units is None:
        K = 1
        bar = 1
    else:
        K = units.kelvin
        bar = units.bar
    if T is None:
        T = 298.15 * K
    if P is None:
        P = 1 * bar
    if U is None:
        U = (
            3.4279e2,
            -5.0866e-3 / K,
            9.4690e-7 / K ** 2,
            -2.0525,
            3.1159e3 * K,
            -1.8289e2 * K,
            -8.0325e3 * bar,
            4.2142e6 * K * bar,
            2.1417 / K * bar,
        )
    if just_return_U:
        return U
    T0 = 273.15 * K
    if warn:
        if _any(T < T0) or _any(T > T0 + 350 * K):
            warnings.warn("Outside temperature range (0-350 degC)")
        else:
            if _any(T > T0 + 70 * K):
                if _any(P > 2000 * bar):
                    warnings.warn("Outside pressure range (2000 bar)")
                else:
                    if _any(P > 5000 * bar):
                        warnings.warn("Outside pressure range (5000 bar)")
    B = U[6] + U[7] / T + U[8] * T
    C = U[3] + U[4] / (U[5] + T)
    eps1000 = U[0] * be.exp(U[1] * T + U[2] * T ** 2)
    return eps1000 + C * be.log((B + P) / (B + 1000.0 * bar))


# bibtex format (generated at doi2bib.org):
reference = {
    "doi": "10.1021/j100475a009",
    "url": "http://dx.doi.org/10.1021/j100475a009",
    "year ": 1979,
    "month": "jun",
    "publisher": "American Chemical Society ({ACS})",
    "volume": 83,
    "number": 12,
    "pages": (1599, 1603),
    "author": "Daniel J. Bradley and Kenneth S. Pitzer",
    "title": (
        "Thermodynamics of electrolytes. 12. Dielectric properties of"
        " water and Debye-Hueckel parameters to 350 C and 1 kbar"
    ),
    "journal": "J. Phys. Chem.",
}



from chempy import Reaction
from chempy.util.testing import requires


@requires("sympy")
def test_Reaction_string():
    from sympy import S

    r = Reaction(
        {"A": 1, "B": 2},
        {"C": S(3) / 2},
        checks=[chk for chk in Reaction.default_checks if chk != "all_integral"],
    )
    assert r.string() == "A + 2 B -> 3/2 C"
    assert r.string(Reaction_arrow="-->", Reaction_coeff_space="") == "A + 2B --> 3/2C"



# -*- coding: utf-8 -*-

from functools import reduce
from operator import attrgetter, add
import sys
from sympy import nsimplify

import pytest

from ..util.arithmeticdict import ArithmeticDict
from ..util.testing import requires
from ..util.parsing import parsing_library
from ..units import default_units, units_library, to_unitless, allclose
from ..chemistry import (
    equilibrium_quotient,
    Substance,
    Species,
    Reaction,
    Equilibrium,
    balance_stoichiometry,
)

if sys.version_info < (3, 6, 0):

    class ModuleNotFoundError(ImportError):
        pass


@requires("numpy")
def test_equilibrium_quotient():
    assert (
        abs(equilibrium_quotient([2.3, 3.7, 5.1], (-1, -1, 1)) - 5.1 / 2.3 / 3.7)
        < 1e-14
    )


@requires(parsing_library)
def test_Substance():
    s = Substance.from_formula("H+")
    assert s.composition == {0: 1, 1: 1}
    assert s.charge == 1
    assert abs(s.mass - 1.008) < 1e-3
    assert s in {s: 1}
    assert hash(s) != hash(Substance.from_formula("He"))


def test_Substance__2():
    H2O = Substance(
        name="H2O", charge=0, latex_name=r"\mathrm{H_{2}O}", data={"pKa": 14}
    )  # will_be_missing_in='0.8.0', use data=...
    OH_m = Substance(name="OH-", charge=-1, latex_name=r"\mathrm{OH^{-}}")
    assert sorted([OH_m, H2O], key=attrgetter("name")) == [H2O, OH_m]


@requires(parsing_library)
def test_Substance__from_formula():
    H2O = Substance.from_formula("H2O")
    assert H2O.composition == {1: 2, 8: 1}
    assert H2O.latex_name == "H_{2}O"
    assert H2O.unicode_name == u"H₂O"
    assert H2O.html_name == u"H<sub>2</sub>O"


@requires(parsing_library)
def test_Species():
    s = Species.from_formula("H2O")
    assert s.phase_idx == 0
    mapping = {"(aq)": 0, "(s)": 1, "(g)": 2}
    assert Species.from_formula("CO2(g)").phase_idx == 3
    assert Species.from_formula("CO2(g)", mapping).phase_idx == 2
    assert Species.from_formula("CO2(aq)", mapping).phase_idx == 0
    assert Species.from_formula("NaCl(s)").phase_idx == 1
    assert Species.from_formula("NaCl(s)", phase_idx=7).phase_idx == 7
    assert Species.from_formula("CO2(aq)", mapping, phase_idx=7).phase_idx == 7

    uranyl_ads = Species.from_formula("UO2+2(ads)", phases={"(aq)": 0, "(ads)": 1})
    assert uranyl_ads.composition == {0: 2, 92: 1, 8: 2}
    assert uranyl_ads.phase_idx == 1


@requires(parsing_library)
def test_Species_default_phase_none():
    with pytest.raises(ValueError):
        Species.from_formula("CO2(aq)", default_phase_idx=None)


def test_Solute():
    from ..chemistry import Solute
    from ..util.pyutil import ChemPyDeprecationWarning

    with pytest.warns(ChemPyDeprecationWarning):
        w = Solute("H2O")
    assert w.name == "H2O"


def test_Reaction():
    substances = s_Hp, s_OHm, s_H2O = (
        Substance("H+", composition={0: 1, 1: 1}),
        Substance("OH-", composition={0: -1, 1: 1, 8: 1}),
        Substance("H2O", composition={0: 0, 1: 2, 8: 1}),
    )
    substance_names = Hp, OHm, H2O = [s.name for s in substances]
    substance_dict = {n: s for n, s in zip(substance_names, substances)}
    r1 = Reaction({Hp: 1, OHm: 1}, {H2O: 1})
    assert sum(r1.composition_violation(substance_dict)) == 0
    assert r1.composition_violation(substance_dict, ["H+"]) == [0]
    viol, cks = r1.composition_violation(substance_dict, True)
    assert viol == [0] * 3 and sorted(cks) == [0, 1, 8]
    assert r1.charge_neutrality_violation(substance_dict) == 0

    r2 = Reaction({Hp: 1, OHm: 1}, {H2O: 2})
    assert sum(r2.composition_violation(substance_dict)) != 0
    assert r2.charge_neutrality_violation(substance_dict) == 0

    r3 = Reaction({Hp: 2, OHm: 1}, {H2O: 2})
    assert sum(r3.composition_violation(substance_dict)) != 0
    assert r3.charge_neutrality_violation(substance_dict) != 0
    assert r3.keys() == {Hp, OHm, H2O}

    with pytest.raises(ValueError):
        Reaction({Hp: -1, OHm: -1}, {H2O: -1})

    assert r1 == Reaction({"H+", "OH-"}, {"H2O"})

    r4 = Reaction({Hp, OHm}, {H2O}, 7)
    ref = {Hp: -3 * 5 * 7, OHm: -3 * 5 * 7, H2O: 3 * 5 * 7}
    r4.rate({Hp: 5, OHm: 3}) == ref
    r5 = r4.copy()
    assert r5 == r4
    assert r5 != r1

    lhs5, rhs5 = {"H+": 1, "OH-": 1}, {"H2O": 1}
    r5 = Reaction(lhs5, rhs5)
    assert r5.reac == lhs5 and r5.prod == rhs5


def test_Reaction__copy():
    r1 = Reaction({"H2O"}, {"H2O"}, checks=())
    r2 = r1.copy()
    assert r1 == r2
    r2.reac["H2O2"] = r2.reac.pop("H2O")  # 1
    r2.prod["H2O2"] = r2.prod.pop("H2O")  # 1
    assert r1.reac == {"H2O": 1} and r1.prod == {"H2O": 1}


@requires(parsing_library)
def test_Reaction__from_string():
    r = Reaction.from_string("H2O -> H+ + OH-; 1e-4", "H2O H+ OH-".split())
    assert r.reac == {"H2O": 1} and r.prod == {"H+": 1, "OH-": 1}

    with pytest.raises(ValueError):
        Reaction.from_string("H2O -> H+ + OH-; 1e-4", "H2O H OH-".split())

    r2 = Reaction.from_string("H2O -> H+ + OH-; 1e-4; ref='important_paper'")
    assert r2.ref == "important_paper"

    with pytest.raises(ValueError):
        Reaction.from_string("H2O -> H2O")
    Reaction.from_string("H2O -> H2O; None; checks=()")

    with pytest.raises(ValueError):
        Reaction({"H2O": 2}, {"H2O2": 2, "O2": -2})

    r4 = Reaction({"H+": 2, "OH-": 1}, {"H2O": 2}, 42.0)
    assert Reaction.from_string(str(r4), "H+ OH- H2O") == r4
    assert Reaction.from_string(str(r4), None) == r4

    r5 = Reaction.from_string(
        "H2O2 -> 0.5 O2 + H2O",
        checks=[c for c in Reaction.default_checks if c != "all_integral"],
    )
    r6 = r5.copy()
    assert r5 == r6

    r7 = Reaction.from_string(
        "H2O -> H + OH; None; data=dict(ref='foo; bar; baz;')  # foobar"
    )
    assert r7.data["ref"] == "foo; bar; baz;"


@requires(parsing_library, units_library)
def test_Reaction_from_string__units():
    r5 = Reaction.from_string("2 H2O2 -> O2 + 2 H2O; 1e-7/molar/second", "H2O O2 H2O2")
    assert to_unitless(r5.param, 1 / default_units.molar / default_units.second) == 1e-7
    r6 = Reaction.from_string("->", checks=())
    assert r6.reac == {} and r6.prod == {}

    r7 = Reaction.from_string("2 A -> B; exp(log(2e-3))*metre**3/mol/hour", None)
    assert r7.reac == {"A": 2} and r7.prod == {"B": 1}
    assert allclose(
        r7.param,
        2e-3 * default_units.metre ** 3 / default_units.mol / default_units.hour,
    )

    with pytest.raises(ValueError):
        Reaction.from_string("2 A -> B; 2e-3/hour", None)

    r8 = Reaction.from_string('A -> B; "k"')
    assert r8.rate_expr().args is None
    assert r8.rate_expr().unique_keys == ("k",)
    r9 = Reaction.from_string("A -> B; 42.0")
    assert r9.rate_expr().args == [42.0]
    assert r9.rate_expr().unique_keys is None

    Reaction.from_string("H+ + OH- -> H2O; 1e10/M/s", "H2O H+ OH-".split())
    with pytest.raises(ValueError):
        Reaction.from_string("H2O -> H+ + OH-; 1e-4/M/s", "H2O H+ OH-".split())


@requires(parsing_library, units_library)
def test_Reaction__check_integral():

    Reaction(*balance_stoichiometry({"H2", "O2"}, {"H2O"}))

    class MyInt:
        def __init__(self, __data):
            self.__data = __data

        def __eq__(self, other):
            return self.__data == other

        def __lt__(self, other):
            return self.__data < other

        def __int__(self):
            return self.__data

        def __rsub__(self, other):
            return other - self.__data

    class MyOne:
        def __eq__(self, other):
            return 1 == other

        def __lt__(self, other):
            return 1 < other

        def __int__(self):
            return 1

        def __rsub__(self, other):
            return other - 1

    Reaction({"H2": MyInt(2), "O2": MyOne()}, {"H2O": 2})

    with pytest.raises(ValueError):
        Reaction({"H2": 1, "O2": 0.5}, {"H2O": 1})


@requires(parsing_library, units_library)
def test_Substance__molar_mass():
    mw_water = Substance.from_formula("H2O").molar_mass(default_units)
    q = mw_water / ((15.9994 + 2 * 1.008) * default_units.gram / default_units.mol)
    assert abs(q - 1) < 1e-3


@requires(units_library)
def test_Equilibrium__as_reactions():
    s = default_units.second
    M = default_units.molar
    H2O, Hp, OHm = map(Substance, "H2O H+ OH-".split())
    eq = Equilibrium({"H2O": 1}, {"H+": 1, "OH-": 1}, 1e-14)
    rate = 1.31e11 / M / s
    fw, bw = eq.as_reactions(kb=rate, units=default_units)
    assert abs((bw.param - rate) / rate) < 1e-15
    assert abs((fw.param / M) / bw.param - 1e-14) / 1e-14 < 1e-15


@requires(parsing_library)
def test_Reaction__latex():
    keys = "H2O H2 O2".split()
    subst = {k: Substance.from_formula(k) for k in keys}
    r2 = Reaction.from_string("2 H2O -> 2 H2 + O2", subst)
    assert r2.latex(subst) == r"2 H_{2}O \rightarrow 2 H_{2} + O_{2}"
    r3 = Reaction.from_string("2 H2O -> 2 H2 + O2; 42; name='split'", subst)
    assert (
        r3.latex(subst, with_param=True, with_name=True)
        == r"2 H_{2}O \rightarrow 2 H_{2} + O_{2}; 42; split"
    )
    assert (
        r3.latex(subst, with_name=True)
        == r"2 H_{2}O \rightarrow 2 H_{2} + O_{2}; split"
    )
    assert (
        r3.latex(subst, with_param=True) == r"2 H_{2}O \rightarrow 2 H_{2} + O_{2}; 42"
    )
    assert r3.latex(subst) == r"2 H_{2}O \rightarrow 2 H_{2} + O_{2}"


@requires(parsing_library)
def test_Reaction__unicode():
    keys = u"H2O H2 O2".split()
    subst = {k: Substance.from_formula(k) for k in keys}
    r2 = Reaction.from_string("2 H2O -> 2 H2 + O2", subst)
    assert r2.unicode(subst) == u"2 H₂O → 2 H₂ + O₂"
    r3 = Reaction.from_string("2 H2O -> 2 H2 + O2; 42; name='split'", subst)
    assert r3.unicode(subst) == u"2 H₂O → 2 H₂ + O₂"
    assert r3.unicode(subst, with_name=True) == u"2 H₂O → 2 H₂ + O₂; split"
    assert (
        r3.unicode(subst, with_name=True, with_param=True)
        == u"2 H₂O → 2 H₂ + O₂; 42; split"
    )
    assert r3.unicode(subst, with_param=True) == u"2 H₂O → 2 H₂ + O₂; 42"


@requires(parsing_library)
def test_Reaction__html():
    keys = "H2O H2 O2".split()
    subst = {k: Substance.from_formula(k) for k in keys}
    r2 = Reaction.from_string("2 H2O -> 2 H2 + O2", subst)
    assert r2.html(subst) == "2 H<sub>2</sub>O &rarr; 2 H<sub>2</sub> + O<sub>2</sub>"
    assert (
        r2.html(subst, Reaction_coeff_fmt=lambda s: "<b>{0}</b>".format(s))
        == "<b>2</b> H<sub>2</sub>O &rarr; <b>2</b> H<sub>2</sub> + O<sub>2</sub>"
    )
    assert (
        r2.html(subst, Reaction_formula_fmt=lambda s: "<b>{0}</b>".format(s))
        == "2 <b>H<sub>2</sub>O</b> &rarr; 2 <b>H<sub>2</sub></b> + <b>O<sub>2</sub></b>"
    )


def test_Reaction__idempotency():
    with pytest.raises(ValueError):
        Reaction({"A": 1}, {"A": 1})
    with pytest.raises(ValueError):
        Reaction({}, {})
    with pytest.raises(ValueError):
        Reaction({"A": 1}, {"B": 1}, inact_reac={"B": 1}, inact_prod={"A": 1})


@requires("sympy")
def test_Equilibrium__eliminate():
    e1 = Equilibrium({"A": 1, "B": 2}, {"C": 3})
    e2 = Equilibrium({"D": 5, "B": 7}, {"E": 11})
    coeff = Equilibrium.eliminate([e1, e2], "B")
    assert coeff == [7, -2]

    e3 = coeff[0] * e1 + coeff[1] * e2
    assert e3.net_stoich("B") == (0,)

    e4 = e1 * coeff[0] + coeff[1] * e2
    assert e4.net_stoich("B") == (0,)

    assert (-e1).reac == {"C": 3}
    assert (e2 * -3).reac == {"E": 33}


@requires(parsing_library, units_library)
def test_Equilibrium__from_string():
    assert Equilibrium.from_string("H2O = H+ + OH-").param is None
    assert Equilibrium.from_string("H2O = H+ + OH-; 1e-14").param == 1e-14
    assert Equilibrium.from_string("H2O = H+ + OH-; 1e-14*molar").param ** 0 == 1
    with pytest.raises(ValueError):
        Equilibrium.from_string("H+ + OH- = H2O; 1e-14*molar")
    eq5 = Equilibrium.from_string(
        "CO2(aq) = CO2(g);" "chempy.henry.HenryWithUnits(3.3e-4 * molar / Pa, 2400 * K)"
    )
    assert eq5.reac == {"CO2(aq)": 1}


def test_Equilibrium__cancel():
    # 2B + C -> E
    e1 = Equilibrium({"A": 26, "B": 20, "C": 7}, {"D": 4, "E": 7})
    e2 = Equilibrium({"A": 13, "B": 3}, {"D": 2})
    coeff = e1.cancel(e2)
    assert coeff == -2


@requires("sympy")
def test_balance_stoichiometry():
    # 4 NH4ClO4 -> 2 N2 + 4 HCl + 6H2O + 5O2
    # 4 Al + 3O2 -> 2Al2O3
    # ---------------------------------------
    # 6 NH4ClO4 + 10 Al + -> 3 N2 + 6 HCl + 9 H2O + 5 Al2O3
    reac, prod = balance_stoichiometry({"NH4ClO4", "Al"}, {"Al2O3", "HCl", "H2O", "N2"})
    assert reac == {"NH4ClO4": 6, "Al": 10}
    assert prod == {"Al2O3": 5, "HCl": 6, "H2O": 9, "N2": 3}

    r3, p3 = balance_stoichiometry({"C2H6", "O2"}, {"H2O", "CO2"})
    assert r3 == {"C2H6": 2, "O2": 7}
    assert p3 == {"CO2": 4, "H2O": 6}

    r4, p4 = balance_stoichiometry({"C7H5(NO2)3", "NH4NO3"}, {"CO", "H2O", "N2"})
    assert r4 == {"C7H5(NO2)3": 2, "NH4NO3": 7}
    assert p4 == {"CO": 14, "H2O": 19, "N2": 10}

    a5, b5 = {"C3H5NO", "CH4", "NH3", "H2O"}, {"C2H6", "CH4O", "CH5N", "CH3N"}
    formulas = list(set.union(a5, b5))
    substances = dict(zip(formulas, map(Substance.from_formula, formulas)))
    compositions = {k: ArithmeticDict(int, substances[k].composition) for k in formulas}
    r5, p5 = balance_stoichiometry(a5, b5)
    compo_reac = nsimplify(
        dict(reduce(add, [compositions[k] * v for k, v in r5.items()]))
    )
    compo_prod = nsimplify(
        dict(reduce(add, [compositions[k] * v for k, v in p5.items()]))
    )
    assert compo_reac == compo_prod

    a6, b6 = map(
        lambda x: set(x.split()), "CuSCN KIO3 HCl;CuSO4 KCl HCN ICl H2O".split(";")
    )
    r6, p6 = balance_stoichiometry(a6, b6)
    assert r6 == dict(CuSCN=4, KIO3=7, HCl=14)
    assert p6 == dict(CuSO4=4, KCl=7, HCN=4, ICl=7, H2O=5)

    r7, p7 = balance_stoichiometry({"Zn+2", "e-"}, {"Zn"})
    assert r7 == {"Zn+2": 1, "e-": 2}
    assert p7 == {"Zn": 1}

    r8, p8 = balance_stoichiometry({"Zn"}, {"Zn+2", "e-"})
    assert r8 == {"Zn": 1}
    assert p8 == {"Zn+2": 1, "e-": 2}


@requires("sympy")
def test_balance_stoichiometry__ordering():
    reac, prod = "CuSCN KIO3 HCl".split(), "CuSO4 KCl HCN ICl H2O".split()
    rxn = Reaction(*balance_stoichiometry(reac, prod))
    res = rxn.string()
    ref = "4 CuSCN + 7 KIO3 + 14 HCl -> 4 CuSO4 + 7 KCl + 4 HCN + 7 ICl + 5 H2O"
    assert res == ref


@requires("sympy")
def test_balance_stoichiometry__simple():
    # 4 NH4ClO4 -> 2 N2 + 4 HCl + 6H2O + 5O2
    # 4 Al + 3O2 -> 2Al2O3
    # ---------------------------------------
    # 6 NH4ClO4 + 10 Al + -> 3 N2 + 6 HCl + 9 H2O + 5 Al2O3
    reac, prod = balance_stoichiometry({"NH4ClO4", "Al"}, {"Al2O3", "HCl", "H2O", "N2"})
    assert reac == {"NH4ClO4": 6, "Al": 10}
    assert prod == {"Al2O3": 5, "HCl": 6, "H2O": 9, "N2": 3}

    r2, p2 = balance_stoichiometry({"Na2CO3"}, {"Na2O", "CO2"})
    assert r2 == {"Na2CO3": 1}
    assert p2 == {"Na2O": 1, "CO2": 1}


@requires("sympy", "pulp")
@pytest.mark.parametrize("underdet", [False, None, True])
def test_balance_stoichiometry__impossible(underdet):
    try:
        from pulp import PulpSolverError
    except ModuleNotFoundError:
        from pulp.solvers import PulpSolverError  # older version of PuLP
    with pytest.raises((ValueError, PulpSolverError)):
        r1, p1 = balance_stoichiometry({"CO"}, {"CO2"}, underdetermined=underdet)


@requires("sympy", "pulp")
def test_balance_stoichiometry__underdetermined():
    try:
        from pulp import PulpSolverError
    except ModuleNotFoundError:
        from pulp.solvers import PulpSolverError  # older version of PuLP

    with pytest.raises(ValueError):
        balance_stoichiometry(
            {"C2H6", "O2"}, {"H2O", "CO2", "CO"}, underdetermined=False
        )
    reac, prod = balance_stoichiometry({"C2H6", "O2"}, {"H2O", "CO2", "CO"})

    r1 = {"C7H5O3-", "O2", "C21H27N7O14P2-2", "H+"}
    p1 = {
        "C7H5O4-",
        "C21H26N7O14P2-",
        "H2O",
    }  # see https://github.com/bjodah/chempy/issues/67
    bal1 = balance_stoichiometry(r1, p1, underdetermined=None)
    assert bal1 == (
        {"C21H27N7O14P2-2": 1, "H+": 1, "C7H5O3-": 1, "O2": 1},
        {"C21H26N7O14P2-": 1, "H2O": 1, "C7H5O4-": 1},
    )

    with pytest.raises(ValueError):
        balance_stoichiometry({"C3H4O3", "H3PO4"}, {"C3H6O3"}, underdetermined=None)

    for underdet in [False, True, None]:
        with pytest.raises((ValueError, PulpSolverError)):
            balance_stoichiometry({"C3H6O3"}, {"C3H4O3"}, underdetermined=underdet)

    with pytest.raises(
        ValueError
    ):  # https://github.com/bjodah/chempy/pull/86#issuecomment-375421609
        balance_stoichiometry(
            {"C21H36N7O16P3S", "C3H4O3"}, {"H2O", "C5H8O3", "C24H38N7O18P3S"}
        )


@requires("sympy", "pulp")
def test_balance_stoichiometry__very_underdetermined():
    r3 = set("O2 Fe Al Cr".split())
    p3 = set("FeO Fe2O3 Fe3O4 Al2O3 Cr2O3 CrO3".split())
    bal3 = balance_stoichiometry(r3, p3, underdetermined=None)
    ref3 = {"Fe": 7, "Al": 2, "Cr": 3, "O2": 9}, {k: 2 if k == "FeO" else 1 for k in p3}
    substances = {k: Substance.from_formula(k) for k in r3 | p3}
    assert all(viol == 0 for viol in Reaction(*ref3).composition_violation(substances))
    assert sum(bal3[0].values()) + sum(bal3[1].values()) <= sum(ref3[0].values()) + sum(
        ref3[1].values()
    )
    assert bal3 == ref3


@requires("sympy", "pulp")
def test_balance_stoichiometry__underdetermined__canoncial():
    # This tests for canonical representation of the underdetermined system
    # where all coefficients are integer and >= 1. It is however of limited
    # practical use (and hence marked ``xfail``) since underdetermined systems
    # have infinite number of solutions. It should however be possible to rewrite
    # the logic so that such canonical results are returned from balance_stoichiometry
    r2 = {"O2", "O3", "C", "NO", "N2O", "NO2", "N2O4"}
    p2 = {"CO", "CO2", "N2"}
    bal2 = balance_stoichiometry(r2, p2, underdetermined=None)
    ref2 = (
        {"O2": 1, "O3": 1, "C": 7, "NO": 1, "N2O": 1, "NO2": 1, "N2O4": 1},
        {"CO": 1, "CO2": 6, "N2": 3},
    )
    substances = {k: Substance.from_formula(k) for k in r2 | p2}
    assert all(viol == 0 for viol in Reaction(*ref2).composition_violation(substances))
    assert sum(bal2[0].values()) + sum(bal2[1].values()) <= sum(ref2[0].values()) + sum(
        ref2[1].values()
    )
    assert bal2 == ref2


@requires("sympy", "pulp")
def test_balance_stoichiometry__substances__underdetermined():
    substances = {
        s.name: s
        for s in [
            Substance("eggs_6pack", composition=dict(eggs=6)),
            Substance("milk_carton", composition=dict(cups_of_milk=4)),
            Substance("flour_bag", composition=dict(spoons_of_flour=30)),
            Substance(
                "pancake", composition=dict(eggs=1, cups_of_milk=1, spoons_of_flour=2)
            ),
            Substance(
                "waffle", composition=dict(eggs=2, cups_of_milk=2, spoons_of_flour=3)
            ),
        ]
    }
    ur1 = {"eggs_6pack", "milk_carton", "flour_bag"}
    up1 = {"pancake", "waffle"}
    br1, bp1 = balance_stoichiometry(
        ur1, up1, substances=substances, underdetermined=None
    )
    ref_r1 = {"eggs_6pack": 6, "flour_bag": 2, "milk_carton": 9}
    ref_p1 = {"pancake": 12, "waffle": 12}
    assert all(
        viol == 0 for viol in Reaction(ref_r1, ref_p1).composition_violation(substances)
    )
    assert all(v > 0 for v in br1.values()) and all(v > 0 for v in bp1.values())
    assert bp1 == ref_p1
    assert br1 == ref_r1


@requires("sympy")
def test_balance_stoichiometry__missing_product_atom():
    with pytest.raises(ValueError):  # No Al on product side
        balance_stoichiometry({"C7H5(NO2)3", "Al", "NH4NO3"}, {"CO", "H2O", "N2"})


@requires("sympy")
def test_balance_stoichiometry__duplicates():
    cases = """
C + CO + CO2 -> C + CO        # suggested solution:  C +      CO2 ->     2 CO
C + CO + CO2 -> C +      CO2  # suggested solution:      2 CO      -> C +      CO2
C + CO + CO2 ->     CO + CO2  # suggested solution:  C +      CO2 ->     2 CO
C + CO       -> C + CO + CO2  # suggested solution:      2 CO      -> C +      CO2
C +      CO2 -> C + CO + CO2  # suggested solution:  C +      CO2 ->     2 CO
    CO + CO2 -> C + CO + CO2  # suggested solution:      2 CO      -> C +      CO2
"""
    for prob, sol in [line.split("#") for line in cases.strip().splitlines()]:
        tst_r = Reaction.from_string(prob)
        ref_r = Reaction.from_string(sol.split(":")[1])
        tst_bal = balance_stoichiometry(
            tst_r.reac, tst_r.prod, allow_duplicates=True, underdetermined=None
        )
        assert Reaction(*tst_bal) == ref_r

    with pytest.raises(ValueError):
        balance_stoichiometry(
            {"C", "CO", "CO2"},
            {"C", "CO", "CO2"},
            allow_duplicates=True,
            underdetermined=None,
        )

    gh120 = {"H4P2O7", "HPO3", "H2O"}, {"H4P2O7", "HPO3"}
    bal120 = balance_stoichiometry(*gh120, allow_duplicates=True, underdetermined=None)
    assert bal120 == ({"HPO3": 2, "H2O": 1}, {"H4P2O7": 1})

    with pytest.raises(ValueError):
        balance_stoichiometry(*gh120)

    # https://github.com/bjodah/chempy/issues/120#issuecomment-434453703
    bal_Mn = balance_stoichiometry(
        {"H2O2", "Mn1", "H1"},
        {"Mn1", "H2O1"},
        allow_duplicates=True,
        underdetermined=None,
    )
    assert bal_Mn == ({"H2O2": 1, "H1": 2}, {"H2O1": 2})

    bal_Mn_COx = balance_stoichiometry(
        {"C", "CO", "CO2", "Mn"},
        {"C", "CO2", "Mn"},
        allow_duplicates=True,
        underdetermined=None,
    )
    assert bal_Mn_COx == ({"CO": 2}, {"C": 1, "CO2": 1})



# -*- coding: utf-8 -*-

try:
    import numpy as np
except ImportError:
    np = None


def broadcast_stack(*args, **kwargs):
    as_scalars = kwargs.pop("as_scalars", False)
    if kwargs != {}:
        raise ValueError("Got unknown kwargs: %s" % kwargs)
    args = [np.atleast_1d(arg) for arg in args]
    if as_scalars:
        args = [arg.reshape(arg.shape + (1,)) if arg.size > 1 else arg for arg in args]
    if all([arg.ndim == 1 for arg in args]):
        return np.concatenate(args)
    head_shape = ()
    leading_length = 0
    for arg in args:
        leading_length += arg.shape[-1]
        if arg.ndim > 1:
            if head_shape == ():
                head_shape = arg.shape[:-1]
            else:
                if arg.shape[:-1] != head_shape:
                    raise ValueError("Incompatible shapes")
    out = np.empty(head_shape + (leading_length,))
    for idx, arg in enumerate(args):
        if arg.shape[-1] != 1:
            raise ValueError("Trailing dimensions needs to be 1")
        out[..., idx] = arg[..., 0]
    return out



# -*- coding: utf-8 -*-


from chempy.chemistry import Species, Equilibrium
from chempy.equilibria import EqSystem


def get_ammonical_cupric_eqsys():
    NH3_complexes = ["CuNH3+2", "Cu(NH3)2+2", "Cu(NH3)3+2", "Cu(NH3)4+2", "Cu(NH3)5+2"]
    OH_complexes = ["Cu2(OH)2+2", "Cu(OH)3-", "Cu(OH)4-2"]
    substances = [
        Species.from_formula(n)
        for n in ["H+", "OH-", "NH4+", "NH3", "H2O", "Cu+2"]
        + NH3_complexes
        + OH_complexes
        + ["Cu(OH)2(s)"]
    ]

    (
        Hp,
        OHm,
        NH4p,
        NH3,
        H2O,
        Cupp,
        CuNH31pp,
        CuNH32pp,
        CuNH33pp,
        CuNH34pp,
        CuNH35pp,
        Cu2OH2pp,
        CuOH3m,
        CuOH4mm,
        CuOH2,
    ) = [s.name for s in substances]
    substances[-1].phase_idx = 1
    init_conc = {
        Hp: 1e-7,
        OHm: 1e-7,
        NH4p: 0,
        NH3: 1.0,
        Cupp: 1e-2,
        CuNH31pp: 0,
        CuNH32pp: 0,
        CuNH33pp: 0,
        CuNH34pp: 0,
        CuNH35pp: 0,
        H2O: 55.5,
        Cu2OH2pp: 0,
        CuOH2: 0,
        CuOH3m: 0,
        CuOH4mm: 0,
    }
    H2O_c = init_conc[H2O]
    w_autop = Equilibrium({H2O: 1}, {Hp: 1, OHm: 1}, 10 ** -14 / H2O_c)
    NH4p_pr = Equilibrium({NH4p: 1}, {Hp: 1, NH3: 1}, 10 ** -9.26)
    CuOH2_s = Equilibrium({CuOH2: 1}, {Cupp: 1, OHm: 2}, 10 ** -18.8)
    CuOH_B3 = Equilibrium({CuOH2: 1, OHm: 1}, {CuOH3m: 1}, 10 ** -3.6)
    CuOH_B4 = Equilibrium({CuOH2: 1, OHm: 2}, {CuOH4mm: 1}, 10 ** -2.7)
    Cu2OH2 = Equilibrium(
        {Cupp: 2, H2O: 2}, {Cu2OH2pp: 1, Hp: 2}, 10 ** -10.6 / H2O_c ** 2
    )
    CuNH3_B1 = Equilibrium({CuNH31pp: 1}, {Cupp: 1, NH3: 1}, 10 ** -4.3)
    CuNH3_B2 = Equilibrium({CuNH32pp: 1}, {Cupp: 1, NH3: 2}, 10 ** -7.9)
    CuNH3_B3 = Equilibrium({CuNH33pp: 1}, {Cupp: 1, NH3: 3}, 10 ** -10.8)
    CuNH3_B4 = Equilibrium({CuNH34pp: 1}, {Cupp: 1, NH3: 4}, 10 ** -13.0)
    CuNH3_B5 = Equilibrium({CuNH35pp: 1}, {Cupp: 1, NH3: 5}, 10 ** -12.4)
    equilibria = (
        w_autop,
        NH4p_pr,
        CuNH3_B1,
        CuNH3_B2,
        CuNH3_B3,
        CuNH3_B4,
        CuNH3_B5,
        Cu2OH2,
        CuOH_B3,
        CuOH_B4,
        CuOH2_s,
    )

    new_eqs = CuOH2_s - CuOH_B3, CuOH2_s - CuOH_B4
    skip_subs, skip_eq = (1, 3)
    simpl_subs = substances[:-skip_subs]
    simpl_eq = equilibria[:-skip_eq] + new_eqs
    simpl_c0 = {k.name: init_conc[k.name] for k in substances[:-skip_subs]}
    return EqSystem(simpl_eq, simpl_subs), simpl_c0



# -*- coding: utf-8 -*-
"""
Utilities for plotting with `bokeh <https://bokeh.pydata.org>`_.
"""

from collections import OrderedDict, defaultdict
from itertools import chain

from chempy.kinetics.ode import get_odesys
from chempy.units import to_unitless, linspace, logspace_from_lin


def integration_with_sliders(
    rsys,
    tend,
    c0,
    parameters,
    fig_kwargs=None,
    slider_kwargs=None,
    conc_bounds=None,
    x_axis_type="linear",
    y_axis_type="linear",
    integrate_kwargs=None,
    odesys_extra=None,
    get_odesys_kw=None,
    integrate=None,
):
    """
    Parameters
    ----------
    rsys : ReactionSystem
    tend : float like
    c0 : dict
        Initial concentrations.
    parameters : dict
        Parameter values.
    fig_kwargs : dict
        Keyword-arguments passed to bokeh's ``Figure``.
    slider_kwargs : dict
        Keyword-arguments passed to bokeh's ``Slider``.
    conc_bounds : dict of dicts
        Mapping substance key to dict of bounds ('start', 'end', 'step').
    x_axis_type : str
    y_axis_type : str
    integrate_kwargs : dict
        Keyword-arguments passed to integrate.
    odesys_extra : tuple
        If odesys & extra have already been generated (avoids call to ``get_odesys``).
    get_odesys_kw : dict
        Keyword-arguments passed to ``get_odesys``.
    integrate : callback
        Defaults to ``odesys.integrate``.

    """

    import numpy as np
    from bokeh.plotting import Figure
    from bokeh.models import ColumnDataSource, Column, Row
    from bokeh.models.widgets import Slider

    if slider_kwargs is None:
        slider_kwargs = {}
    if get_odesys_kw is None:
        get_odesys_kw = {}
    if odesys_extra is None:
        odesys, extra = get_odesys(rsys, **get_odesys_kw)
    else:
        odesys, extra = odesys_extra
    if integrate is None:
        integrate = odesys.integrate

    state_keys, rarg_keys, p_units = [
        extra[k] for k in ("param_keys", "unique", "p_units")
    ]
    output_conc_unit = get_odesys_kw.get("output_conc_unit", None)
    output_time_unit = get_odesys_kw.get("output_time_unit", None)
    unit_registry = get_odesys_kw.get("unit_registry", None)
    if output_conc_unit is None:
        if unit_registry is not None:
            raise ValueError(
                "if unit_registry is given, output_conc_unit must also be given"
            )
        output_conc_unit = 1
    if output_time_unit is None:
        if unit_registry is not None:
            raise ValueError(
                "if unit_registry is given, output_time_unit must also be given"
            )
        output_conc_unit = 1

    param_keys = list(chain(state_keys, rarg_keys))
    if x_axis_type == "linear":
        tout = linspace(tend * 0, tend)
    elif x_axis_type == "log":
        tout = logspace_from_lin(tend * 1e-9, tend)
    else:
        raise NotImplementedError("Unknown x_axis_type: %s" % x_axis_type)

    result = integrate(tout, c0, parameters, **(integrate_kwargs or {}))
    sources = [
        ColumnDataSource(
            data={
                "tout": to_unitless(result.xout, output_time_unit),
                k: to_unitless(result.yout[:, idx], output_conc_unit),
            }
        )
        for idx, k in enumerate(rsys.substances)
    ]
    if fig_kwargs is None:
        Cmax = np.max(result.yout)
        x_range = list(to_unitless([result.xout[0], result.xout[-1]], output_time_unit))
        y_range = list(to_unitless([Cmax * 0, Cmax * 1.1], output_conc_unit))
        fig_kwargs = dict(
            plot_height=400,
            plot_width=400,
            title="C vs t",
            tools="crosshair,pan,reset,save,wheel_zoom",
            x_range=x_range,
            y_range=y_range,
            x_axis_type=x_axis_type,
            y_axis_type=y_axis_type,
        )
    plot = Figure(**fig_kwargs)

    colors = "red green blue black cyan magenta".split()
    for idx, k in enumerate(rsys.substances):
        plot.line(
            "tout",
            k,
            source=sources[idx],
            line_width=3,
            line_alpha=0.6,
            color=colors[idx % len(colors)],
        )

    def _C(k):
        return to_unitless(c0[k], output_conc_unit)

    if p_units is None:
        p_units = [None] * len(param_keys)
    p_ul = [to_unitless(parameters[k], _u) for k, _u in zip(param_keys, p_units)]

    def _dict_to_unitless(d, u):
        return {k: to_unitless(v, u) for k, v in d.items()}

    c0_widgets = OrderedDict()
    for k in rsys.substances:
        if conc_bounds is not None and k in conc_bounds:
            if k in slider_kwargs:
                raise ValueError("Key '%s' both in slider_kwargs and conc_bounds" % k)
            slider_defaults = _dict_to_unitless(conc_bounds[k], output_conc_unit)
        else:
            ck = _C(k)
            if ck == 0:
                max_ = max(*[_C(k) for k in rsys.substances])
                slider_defaults = dict(start=0, end=max_, step=max_ / 100)
            else:
                slider_defaults = dict(start=_C(k) / 2, end=_C(k) * 2, step=_C(k) / 10)
        c0_widgets[k] = Slider(
            title=(k + " / " + output_conc_unit.dimensionality.unicode)
            if hasattr(output_conc_unit, "dimensionality")
            else k,
            value=_C(k),
            **slider_kwargs.get(k, slider_defaults)
        )

    param_widgets = OrderedDict(
        [
            (
                k,
                Slider(
                    title=k if u is None else k + " / " + u.dimensionality.unicode,
                    value=v,
                    **_dict_to_unitless(
                        slider_kwargs.get(
                            k, dict(start=v / 10, end=v * 10, step=v / 10)
                        ),
                        u,
                    )
                ),
            )
            for k, v, u in zip(param_keys, p_ul, p_units)
        ]
    )
    all_widgets = list(chain(param_widgets.values(), c0_widgets.values()))

    def update_data(attrname, old, new):
        _c0 = defaultdict(lambda: 0 * output_conc_unit)
        for k, w in c0_widgets.items():
            _c0[k] = w.value * output_conc_unit
        _params = {}
        for (k, w), u in zip(param_widgets.items(), p_units):
            _params[k] = w.value if u is None else w.value * u
        _result = integrate(tout, _c0, _params)
        for idx, k in enumerate(rsys.substances):
            sources[idx].data = {
                "tout": to_unitless(_result.xout, output_time_unit),
                k: to_unitless(_result.yout[:, idx], output_conc_unit),
            }

    for w in all_widgets:
        w.on_change("value", update_data)

    inputs = Column(children=all_widgets)
    return Row(children=[inputs, plot], width=800)



# -*- coding: utf-8 -*-
"""
"Sandbox" module for exploring API useful for digital labbooks.

Examples
--------
>>> from chempy.units import to_unitless, default_units as u
>>> s1 = Solution(0.1*u.dm3, {'CH3OH': 0.1 * u.molar})
>>> s2 = Solution(0.3*u.dm3, {'CH3OH': 0.4 * u.molar, 'Na+': 2e-3*u.molar, 'Cl-': 2e-3*u.molar})
>>> s3 = s1 + s2
>>> bool(abs(to_unitless(s3.volume - 4e-4 * u.m**3, u.dm3)) < 1e-15)
True
>>> s3.concentrations.isclose({'CH3OH': 0.325*u.molar, 'Na+': 1.5e-3*u.molar, 'Cl-': 1.5e-3*u.molar})
True
>>> s4 = s3.dissolve({'CH3OH': 1*u.gram})
>>> bool(abs(s4.concentrations['CH3OH'] - (0.325 + 1/(12.011 + 4*1.008 + 15.999)/.4)*u.molar) < 1e-4)
True

"""

import copy

from .chemistry import Substance
from .units import (
    get_derived_unit,
    html_of_unit,
    is_unitless,
    SI_base_registry,
    to_unitless,
    rescale,
    default_units as u,
)
from .util.arithmeticdict import ArithmeticDict, _imul, _itruediv
from .printing import as_per_substance_html_table


class QuantityDict(ArithmeticDict):
    def __init__(self, units, *args, **kwargs):
        self.units = units
        super(QuantityDict, self).__init__(lambda: 0 * self.units, *args, **kwargs)
        self._check()

    @classmethod
    def of_quantity(cls, quantity_name, *args, **kwargs):
        instance = cls(
            get_derived_unit(SI_base_registry, quantity_name), *args, **kwargs
        )
        instance.quantity_name = quantity_name
        return instance

    def rescale(self, new_units):
        return self.__class__(
            new_units, {k: rescale(v, new_units) for k, v in self.items()}
        )

    def _repr_html_(self):
        if hasattr(self, "quantity_name"):
            header = self.quantity_name.capitalize() + " / "
        else:
            header = ""
        header += html_of_unit(self.units)
        tab = as_per_substance_html_table(to_unitless(self, self.units), header=header)
        return tab._repr_html_()

    def _check(self):
        for k, v in self.items():
            if not is_unitless(v / self.units):
                raise ValueError(
                    "entry for %s (%s) is not compatible with %s" % (k, v, self.units)
                )

    def __setitem__(self, key, value):
        if not is_unitless(value / self.units):
            raise ValueError(
                "entry for %s (%s) is not compatible with %s" % (key, value, self.units)
            )
        super(QuantityDict, self).__setitem__(key, value)

    def copy(self):
        return self.__class__(self.units, copy.deepcopy(list(self.items())))

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, repr(self.units), dict(self)
        )

    def __mul__(self, other):
        d = dict(copy.deepcopy(list(self.items())))
        _imul(d, other)
        return self.__class__(self.units * getattr(other, "units", 1), d)

    def __truediv__(self, other):
        d = dict(copy.deepcopy(list(self.items())))
        _itruediv(d, other)
        return self.__class__(self.units / getattr(other, "units", 1), d)

    def __floordiv__(self, other):
        a = self.copy()
        if getattr(other, "units", 1) != 1:
            raise ValueError("Floor division with quantities not defined")
        a //= other
        return a

    def __rtruediv__(self, other):
        """other / self"""
        return self.__class__(
            getattr(other, "units", 1) / self.units,
            {k: other / v for k, v in self.items()},
        )

    def __rfloordiv__(self, other):
        """other // self"""
        return self.__class__(
            getattr(other, "units", 1) / self.units,
            {k: other // v for k, v in self.items()},
        )


class AutoRegisteringSubstanceDict(object):
    def __init__(self, factory=Substance.from_formula):
        self.factory = factory
        self._store = {}

    def __getitem__(self, key):
        if key not in self._store:
            self._store[key] = self.factory(key)
        return self._store[key]


class Solution(object):
    def __init__(self, volume, concentrations, substances=None, solvent=None):
        if not is_unitless(volume / u.dm3):
            raise ValueError("volume need to have a unit (e.g. dm3)")
        self.volume = volume
        self.concentrations = QuantityDict(u.molar, concentrations)
        if substances is None:
            substances = AutoRegisteringSubstanceDict()
        self.substances = substances
        self.solvent = solvent

    def __eq__(self, other):
        if not isinstance(other, Solution):
            return NotImplemented
        return all(
            [
                getattr(self, k) == getattr(other, k)
                for k in "volume concentrations substances solvent".split()
            ]
        )

    def __add__(self, other):
        if self.solvent != other.solvent:
            raise NotImplementedError(
                "Mixed solvent should be represented as concentrations"
            )
        tot_amount = (
            self.concentrations * self.volume + other.concentrations * other.volume
        )
        tot_vol = self.volume + other.volume
        return Solution(tot_vol, tot_amount / tot_vol, self.substances, self.solvent)

    def dissolve(self, masses):
        contrib = QuantityDict(
            u.molar,
            {
                k: v / self.substances[k].molar_mass() / self.volume
                for k, v in masses.items()
            },
        )
        return Solution(
            self.volume, self.concentrations + contrib, self.substances, self.solvent
        )

    def withdraw(self, volume):
        if volume > self.volume:
            raise ValueError(
                "Cannot withdraw a volume greater than the solution volume"
            )
        if volume < volume * 0:
            raise ValueError("Cannot withdraw a negative volume")
        self.volume -= volume
        return Solution(volume, self.concentrations, self.substances, self.solvent)



# -*- coding: utf-8 -*-


from ..henry import Henry, HenryWithUnits
from ..units import units_library, allclose, default_units as u
from ..util.testing import requires


def test_Henry():
    kH_O2 = Henry(1.2e-3, 1800, ref="carpenter_1966")
    assert abs(kH_O2(298.15) - 1.2e-3) < 1e-4
    assert abs(kH_O2.get_c_at_T_and_P(290, 1) - 0.001421892) < 1e-8
    assert abs(kH_O2.get_P_at_T_and_c(310, 1e-3) - 1.05) < 1e-3


@requires(units_library, "numpy")
def test_Henry__with_units():
    import numpy as np

    kH_H2 = HenryWithUnits(7.8e-4 * u.molar / u.atm, 640 * u.K, ref="dean_1992")

    assert allclose(kH_H2.get_kH_at_T(300 * u.K), 7.697430323e-4 * u.molar / u.atm)
    kH = kH_H2.get_c_at_T_and_P(np.linspace(297.5, 298.65, 3) * u.K, 0.1 * u.bar)
    assert allclose(kH, 7.7e-5 * u.molar, rtol=1e-5, atol=1e-6 * u.molar)
    kH = kH_H2.get_P_at_T_and_c(298.15 * u.K, np.linspace(2e-3, 2.1e-3, 3) * u.molar)
    assert allclose(kH, 2.65 * u.atm, rtol=1e-5, atol=0.2 * u.bar)


@requires(units_library)
def test_HenryWithUnits():
    kH_H2 = HenryWithUnits(7.8e-4 * u.molar / u.atm, 640 * u.K, ref="dean_1992")
    Hcp = kH_H2(300 * u.K)
    assert allclose(Hcp, 7.697430323e-4 * u.molar / u.atm)



# -*- coding: utf-8 -*-

from chempy.util.testing import requires
from chempy.units import units_library, default_units as u

from ..numbers import (
    roman,
    _float_str_w_uncert,
    number_to_scientific_html,
    number_to_scientific_latex,
    number_to_scientific_unicode,
)


def test__float_str_w_uncert():
    assert _float_str_w_uncert(-5739, 16.34, 2) == "-5739(16)"
    assert _float_str_w_uncert(-5739, 16.9, 2) == "-5739(17)"
    assert _float_str_w_uncert(0.0123, 0.00169, 2) == "0.0123(17)"
    assert _float_str_w_uncert(0.01234, 0.00169, 2) == "0.0123(17)"
    assert _float_str_w_uncert(0.01234, 0.0016501, 2) == "0.0123(17)"
    assert _float_str_w_uncert(0.01234, 0.0016501, 1) == "0.012(2)"
    assert _float_str_w_uncert(-9.99752e15, 349e10, 2) == "-9.9975(35)e15"
    assert _float_str_w_uncert(-9.99752e5, 349, 2) == "-999750(350)"
    assert _float_str_w_uncert(-9.99752e5, 349, 3) == "-999752(349)"
    assert _float_str_w_uncert(315, 17.9e-4, 2) == "315.0000(18)"


def test_roman():
    assert roman(4) == "IV"
    assert roman(20) == "XX"
    assert roman(94) == "XCIV"
    assert roman(501) == "DI"


def test_number_to_scientific_html():
    assert number_to_scientific_html(2e-17) == "2&sdot;10<sup>-17</sup>"
    assert number_to_scientific_html(1e-17) == "10<sup>-17</sup>"


def test_number_to_scientific_latex():
    assert number_to_scientific_latex(2e-17) == r"2\cdot 10^{-17}"
    assert number_to_scientific_latex(1e-17) == "10^{-17}"
    assert number_to_scientific_latex(315, 17.9e-4, fmt=2) == "315.0000(18)"


@requires(units_library)
def test_number_to_scientific_latex__units():
    assert (
        number_to_scientific_latex(315 * u.km, 17.9 * u.dm, fmt=2)
        == r"315.0000(18)\,\mathrm{km}"
    )
    assert (
        number_to_scientific_latex(315 * u.km, 17.9 * u.dm, u.m, fmt=2)
        == r"315000.0(18)\,\mathrm{m}"
    )
    assert (
        number_to_scientific_latex(1319 * u.km, 41207 * u.m, u.m, fmt=1)
        == r"1.32(4)\cdot 10^{6}\,\mathrm{m}"
    )


def test_number_to_scientific_unicode():
    assert number_to_scientific_unicode(2e-17) == u"2·10⁻¹⁷"
    assert number_to_scientific_unicode(1e-17) == u"10⁻¹⁷"



# -*- coding: utf-8 -*-


import math

from chempy.chemistry import Reaction
from chempy.util.testing import requires
from chempy.units import units_library, allclose, default_units as u

from ..arrhenius import arrhenius_equation, ArrheniusParam, ArrheniusParamWithUnits


def test_arrhenius_equation():
    assert abs(arrhenius_equation(3, 831.4472, 100) - 3 / 2.7182818) < 1e-7


_A1, _Ea1, _T1, _k1 = 1e10, 42e3, 273.15, 1e10 * math.exp(-42e3 / (8.3145 * 273.15))


def test_ArrheniusParam():
    k = ArrheniusParam(_A1, _Ea1)(_T1)
    assert abs((k - _k1) / _k1) < 1e-4


@requires("numpy")
def test_ArrheniusParam__from_rateconst_at_T():
    ap = ArrheniusParam.from_rateconst_at_T(_Ea1, (_T1, _k1))
    assert abs((ap.A - _A1) / _A1) < 1e-4


def _get_ref2_units():
    A__s = 1e10
    act_J__mol = 42e3
    freezing_K = 273.15

    class ValueHolder:
        A = A__s / u.s
        Ea = act_J__mol * u.J / u.mol
        T = freezing_K * u.K
        k = A__s / u.s * math.exp(-act_J__mol / (8.3145 * freezing_K))

    return ValueHolder()


@requires(units_library)
def test_ArrheniusParamWithUnits():
    _2 = _get_ref2_units()
    ap = ArrheniusParamWithUnits(_2.A, _2.Ea)

    k = ap(_2.T)
    assert abs((k - _2.k) / _2.k) < 1e-4

    r = Reaction({"H2O2": 1}, {"OH": 2}, ap)
    ratc = r.rate_expr().rate_coeff({"temperature": _2.T})
    assert allclose(ratc, _2.k, rtol=1e-4)


@requires(units_library)
def test_ArrheniusParamWithUnits__from_rateconst_at_T():
    _2 = _get_ref2_units()
    apu = ArrheniusParamWithUnits.from_rateconst_at_T(_2.Ea, (_2.T, _2.k))
    assert allclose(apu(_2.T), _2.k)
